unit ICEvents;

(*	This file is part of the Internet Configuration system and is placed in the public domain for the benefit of all.

	This module implements the inside of the application's main event loop. It implements the
	bulk of the standard Macintosh "get this event, Handle this event in the default way" stuff,
	and calls out the ICWindows to do the real work.  There is not a lot of excitement here.
*)

interface

	uses
		Events;
		
	procedure HandleOneEvent(event : EventRecord);
		(* Handle one event. Also called by various filter procs. *)

	procedure GetAndHandleAnEvent;
		(* Implements the application's main event loop. *)

implementation

	uses
		EPPC,
		ToolUtils,
		AppleEvents,
		DiskInit,
		MacWindows,
		Devices, 
		{$IFC NOT TARGET_API_MAC_CARBON} ICAccessors, {$ENDC}

		ICDebug,
		ICResources,
		ICGlobals,
		ICMiscSubs,
		ICStringLists,
		ICWindowGlobals,
		ICWindows,
		ICDialogs,
		ICDocument,
		ICInstall,
		ICText;

	procedure DoDiskEvent (message: longint);
		(* Handle bad disk inserted events in the usual way. *)
		var
			pt: Point;
			junkErr: OSStatus;
			screenBitsBounds : Rect;
	begin
		if (hiwrd(message) <> noErr) then begin
			screenBitsBounds := GetScreenBitsBounds;
			pt.h := ((screenBitsBounds.right - screenBitsBounds.left - 304) div 2);
			pt.v := ((screenBitsBounds.bottom - screenBitsBounds.top - 156) div 3);
			InitCursor;
			junkErr := DIBadMount(pt, message);
		end; (* if *)
	end; (* DoDiskEvent *)

	procedure DoActivateDeactivate (window: WindowPtr; activate: Boolean);
		(* Handle activate/deactivate events by highlighting the window appropriately and also
			passing the event along to ICWindows, which does the right thing for our TextEdit fields.
		*)
	begin
		if activate then begin
			SelectWindow(window);
		end; (* if *)
		WindowActivateDeactivate(window, activate);
	end; (* DoActivateDeactivate *)

	procedure DoSuspendResume (resume: Boolean);
		(* Handle suspend and resume events by activating/deactivating the front window. *)
	begin
		gInForeground := resume;
		if FrontWindow <> nil then begin
			DoActivateDeactivate(FrontWindow, resume);
		end; (* if *)
		InitCursor;
	end; (* DoSuspendResume *)
	
	function IsDAWindow(wp : WindowPtr) : Boolean;
	begin
		{$ifc TARGET_API_MAC_CARBON}
			{$unused wp}
			IsDAWindow := false;
		{$elsec}
			IsDAWindow := WindowPeek(wp)^.windowKind < 0;
		{$endc}
	end; (* IsDAWindow *)
	
	function DoCloseWindow(wp : WindowPtr) : OSStatus;
		var
			err : OSStatus;
	begin
		err := noErr;
		if wp <> nil then begin
			if IsDAWindow(wp) then begin
				{$ifc not TARGET_API_MAC_CARBON}
					CloseDeskAcc(WindowPeek(wp)^.windowKind);
				{$endc}
			end else begin
				err := DoCloseDocumentWindow(wp);
			end; (* if *)
		end; (* if *)
		DoCloseWindow := err;
	end; (* DoCloseWindow *)

	procedure AdjustMenus;
		(* Adjust all the menus in preparation for a MenuSelect, MenuKey, or DimMenuTitles operation. *)
		var
			junk : OSStatus;
			editMenu : MenuHandle;
			wp : WindowPtr;
	begin
		(*	
			Push the contents of all whats out to the IC database.  This
			effectively updates the IsProfileDirty result with the latest
			contents of the window, so that the menu commands which depend
			on it are updated appropriately.  There's not a lot we can do
			with the error code at this stage, so we just junk it.
		*)
		
		junk := WindowsFlushAll;
		ICAssert(junk = noErr);

		(* File *)
		SetMenuItemEnable(GetMenuHandle(mFile), iSave, IsProfileDirty);

		SetMenuItemEnable(GetMenuHandle(mFile), iRevert, IsProfileDirty);
		(* Edit *)
		wp := FrontWindow;
		if (wp <> nil) & IsDAWindow(wp) then begin
			(* If we have a DA at the front, turn on the standard edit commands. *)
			editMenu := GetMenuHandle(mEdit);
			SetMenuItemEnable(editMenu, iUndo, false);
			SetMenuItemEnable(editMenu, iCut, true);
			SetMenuItemEnable(editMenu, iCopy, true);
			SetMenuItemEnable(editMenu, iPaste, true);
			SetMenuItemEnable(editMenu, iClear, true);
			SetMenuItemEnable(editMenu, iSelectAll, false);
		end else begin
			WindowsAdjustEditMenu;
		end; (* if *)
		(* Sets *)
		AdjustSetsMenu;
		(* Special *)
		AdjustInstallMenu;
		// The Synchronise Sets item in the Special
		// menu is adjusted in the AdjustSetsMenu.
		(* Windows *)
		SetMenuItemText(GetMenuHandle(mWindows), iMainWindowName, GetProfileName);
	end; (* AdjustMenus *)

	procedure DimMenuTitles;
		(* Called periodically by the main event loop to enable/disable the titles of menus. This routine
			works by first adjusting the menus to reflect their current state, then seeing whether the
			menu title state matches the state of the items, and setting it otherwise. We also keep track
			of whether any of the menu titles have changed to decide whether the menu bar needs redrawing.
		*)
		var
			menuH: MenuHandle;
			i: integer;
			wasTitleEnabled: Boolean;
			shouldTitleBeEnabled: Boolean;
			menuBarNeedRedraw: Boolean;
	begin
		AdjustMenus;

		(* Walk the Edit menu to see if any menu items are set. *)
		menuH := GetMenuHandle(mEdit);
		wasTitleEnabled := IsMenuItemEnabled(menuH, 0);
		shouldTitleBeEnabled := false;
		for i := 1 to CountMenuItems(menuH) do begin
			if IsMenuItemEnabled(menuH, i) then begin
				shouldTitleBeEnabled := true;
				leave;
			end; (* if *)
		end; (* for *)
		SetMenuItemEnable(menuH, 0, shouldTitleBeEnabled);
		menuBarNeedRedraw := (shouldTitleBeEnabled <> wasTitleEnabled);
		
		(* Redraw the menu bar if necessary. *)
		if menuBarNeedRedraw then begin
			DrawMenuBar;
		end; (* if *)
	end; (* DimMenuTitles *)

	procedure DoMenu (menuAndItem : longint);
		(* Handle menu events. Basically this is the standard "Handle the boring events here and dispatch
			the rest to the appropriate place."
		*)
		var
			startTime : longint;
			menu : integer;
			item : integer;
			{$ifc not TARGET_API_MAC_CARBON}
				deskAccessoryName: Str255;
				junk: OSStatus;
			{$endc}
			junkLong: UInt32;
	begin
		startTime := TickCount;
		menu := hiwrd(menuAndItem);
		item := lowrd(menuAndItem);
		case menu of
			mApple:  begin
				case item of
					iAbout: 
						DisplayError(acOpenWindow, WindowOpen(WT_About));
					otherwise begin
						{$ifc not TARGET_API_MAC_CARBON}
							GetMenuItemText(GetMenuHandle(mApple), item, deskAccessoryName);
							junk := OpenDeskAcc(deskAccessoryName);
						{$endc}
					end
				end; (* case *)
			end;
			mFile:  begin
				case item of
					iClose: 
						DisplayError(acCloseWindow, DoCloseWindow(FrontWindow));
					iSave: 
						DisplayError(acSave, DoSaveDocument);
					iRevert:
						DisplayError(acRevert, DoRevertDocument);
					iQuit: 
						DisplayError(acQuit, DoQuit);
					otherwise
						;
				end; (* case *)
			end;
			mEdit:  begin
				case item of
					iUndo, iCut, iCopy, iPaste, iClear, iSelectAll:
						{$ifc TARGET_API_MAC_CARBON}
							if item > iClear then begin
								WindowsDoEditMenu(item);
							end; (* if *)
						{$elsec}
							if (item > iClear) | not SystemEdit(item - 1) then begin
								WindowsDoEditMenu(item);
							end; (* if *)
						{$endc}
					otherwise
						;
				end;
			end;
			mSets :
				DoSetsMenu(item);
			mSpecial:
				case item of
					iInstallExtension,
					iSaveExtension:
						DoInstallMenu(item);
					iSynchroniseSets:
						DoSynchroniseSets;
				end; (* case *)
			mWindows:
				DisplayError(acOpenWindow, WindowOpen(WindowType(ord(WT_Main) + item - 1)));

		{$ifc icDebug}
			mDebug:
				case item of
					iLockAllPreferences:
						begin
							gLockAllPreferences := not gLockAllPreferences;
							CheckMenuItem(GetMenuHandle(mDebug), iLockAllPreferences, gLockAllPreferences);
						end
					iTextMustUseTextEdit:
						begin
							gTextMustUseTextEdit := not gTextMustUseTextEdit;
							CheckMenuItem(GetMenuHandle(mDebug), iTextMustUseTextEdit, gTextMustUseTextEdit);
						end
					iDisableNavServices:
						begin
							gHasNavServices := not gHasNavServices;
							CheckMenuItem(GetMenuHandle(mDebug), iDisableNavServices, not gHasNavServices);
						end
				end; (* case *)
		{$endc}

			otherwise
				;
		end; (* case *)
		if not gQuitNow then begin

			(* For very fast menu commands, hang around until the user has time
				to actually see the menu highlighting.
			*)
			while TickCount < (startTime + 6) do begin
				Delay(1, junkLong);
			end; (* while *)
			
			HiliteMenu(0);
		end; (* if *)
	end; (* DoMenu *)

	procedure DoMainClick (event: EventRecord; wp: WindowPtr; clickCode: integer);
		(* Handle a mouseDown event that hasn't been fielded by the Dialog Manager. Basically
			this translates to clicks in the menu bar and close box, dragging windows, selecting
			non-frontmost windows.
		*)
		var
			screenBitsBounds : Rect;
	begin
		(* Then go and do each type of event. *)
		case clickCode of
			inMenuBar:  begin
				AdjustMenus;
				DoMenu(MenuSelect(event.where));
			end;
			
			inDrag:  begin
				SetPort(GrafPtr(GetWindowPort(wp)));			(* I'm not quite sure what this is doing here. *)
				screenBitsBounds := GetScreenBitsBounds;
				DragWindow(wp, event.where, @screenBitsBounds);
			end;
			
			inGoAway:  begin
				if TrackGoAway(wp, event.where) then begin
					DisplayError(acCloseWindow, DoCloseWindow(wp));
				end; (* if *)
			end;
			
			inContent:  begin
				if (wp <> nil) & (wp <> FrontWindow) then begin
					SelectWindow(wp);
				end; (* if *)
			end;
			
			inSysWindow: 
				{$ifc TARGET_API_MAC_CARBON}
					(* do nothing -- Carbon handles this for us *) ;
				{$elsec}
					SystemClick(event, wp);
				{$endc}

			inGrow: 
				;
			inZoomIn, inZoomOut: 
				;
			otherwise
				(* do nothing *)
		end; (* case *)
	end; (* DoMainClick *)

	procedure HandleOneEvent(event : EventRecord);
		(* Handle one event. Also called by various filter procs. *)
		var
			eventPending: Boolean;
			clickCode: integer;
			wp: WindowPtr;
			dialogHit : DialogPtr;
			menuAndItem: longint;
			junk: OSStatus;
			itemHit: integer;
	begin
		WindowsIdle;
		eventPending := true;
		
		(* Prehandle keyboard events. *)
		if eventPending & ((event.what = keyDown) | (event.what = autoKey)) then begin
			(* First command keys... *)
			if band(event.modifiers, cmdKey) <> 0 then begin
				AdjustMenus;
				menuAndItem := MenuKey(chr(band(event.message, charCodeMask)));
				if menuAndItem <> 0 then begin
					DoMenu(menuAndItem);
					eventPending := false;
				end; (* if *)
			end; (* if *)
			(* ... then prehandle normal keys. *)
			if eventPending & WindowsEarlyHandleKey(event) then begin
				eventPending := false;
			end; (* if *)
		end; (* if *)
		
		(* Hand the event off to the Dialog Manager. *)
		if eventPending then begin

			// We dont give Dialog Manager a crack at keyboard
			// events because we want to do filtering and such inside
			// the Text Whats before we direct them to the right control.
			
			if (event.what <> keyDown) & (event.what <> autoKey) then begin
				if IsDialogEvent(event) then begin
					if DialogSelect(event, dialogHit, itemHit) then begin
						ICAssert(itemHit <> 0);
						WindowItemClick(dialogHit, event, itemHit);
						eventPending := false;
					end; (* if *)
				end; (* if *)
			end; (* if *)
		end; (* if *)
		
		(* Do a classic main event loop case statement. *)
		if eventPending then begin
			case event.what of
				mouseDown:  begin
					clickCode := FindWindow(event.where, wp);
					if wp = nil then begin
						wp := FrontWindow;
					end;
					DoMainClick(event, wp, clickCode);
				end;

				keyDown, autoKey:
					WindowsDoKey(event);

				updateEvt: 
					(* do nothing -- all update events are handled by the Dialog Manager *) ;

				activateEvt: 
					DoActivateDeactivate(WindowPtr(event.message), odd(event.modifiers));

				osEvt: 
					if band(bsr(event.message, 24), $FF) = suspendResumeMessage then begin
						DoSuspendResume(band(event.message, resumeFlag) <> 0);
					end;

				diskEvt: 
					{$ifc TARGET_API_MAC_CARBON}
						(* do nothing -- Carbon handles this for us *) ;
					{$elsec}
						DoDiskEvent(event.message);
					{$endc}

				kHighLevelEvent: 
					junk := AEProcessAppleEvent(event);
				otherwise
			 (* do nothing *)
			end; (* case *)
		end; (* if *)
	end; (* HandleOneEvent *)
	
	procedure GetAndHandleAnEvent;
		(* Get an event and Handle it! *)
		var
			junkBool: Boolean;
			event: EventRecord;
	begin
		DimMenuTitles;
		junkBool := WaitNextEvent(everyEvent, event, 10, nil);
		HandleOneEvent(event);
	end; (* GetAndHandleAnEvent *)

end. (* ICEvents *)