unit ICFSSpecWhats;

(*	This file is part of the Internet Configuration system and is placed in the public domain for the benefit of all.

	This module implements the FSSpec What, which provides a button that
	you click on to set a folder. This seems like a simple exercise, and it
	would be, except for the issues related to tracking the selected
	item  the module has to operate successfully with or without
	Alias Manager, and interoperate with preference files that were
	created with or without Alias Manager.
	
	To understand this module, you really need to understand the
	ICFileSpec data type, as defined in InternetConfig and documented in the
	IC Programming Documentation. The gist of an ICFileSpec is
	that it's a dual role alias, a handle with a very simplistic Poor
	Man's Alias at the front, followed by a standard Alias Manager alias.
	This is required to support System 6 operations )-:
*)

interface

	uses
		ICWindowGlobals;

	(* ***** Standard What Interface ***** *)
	
	(* See the comments in ICWindowGlobals. *)

	function WhatOpenFSSpec (wt: WindowType; item: integer): OSStatus;
	function WhatFlushFSSpec (wt: WindowType; item: integer): OSStatus;
	function WhatClickFSSpec (wt: WindowType; item: integer; event: EventRecord): OSStatus;

implementation

	uses
		Folders,
		ToolUtils,
		Dialogs,
		Aliases,
		StringCompare,
		TextUtils,

		InternetConfig, 
		
		ICDialogs, 
		ICMiscSubs, 
		ICCommonSubs,
		ICWindowGlobals, 
		ICGlobals, 
		ICStandardFile,
		ICPreferenceCache;

	(* ***** Utility Routines ***** *)

	function FindVolumeByNameAndDate (volumeName: Str31; volumeCreationDate: longint;
												var vRefNum: integer): OSStatus;
		(* This routine attempt to find the volume (and return its vRefNum)
			based on its name and creation date. This is part of a Poor Man's
			Alias Manager.
			
			The routine uses a two pass algorithm. In the first pass, it looks
			for a match on both the name and the creation date. On the second
			pass, it'll settle for just a name match.
			
			There are way to many "leave"s in this code. My only excuse
			is that I didn't write it!
		*)
		var
			err: OSStatus;
			pass : (kNameAndDateMatchPass, kNameOnlyMatchPass);
			volumeIndex: integer;
			pb: HParamBlockRec;
			thisVolumesName: Str255;
	begin
		for pass := kNameAndDateMatchPass to kNameOnlyMatchPass do begin
			volumeIndex := 1;
			while true do begin
				pb.ioVolIndex := volumeIndex;
				volumeIndex := volumeIndex + 1;
				pb.ioNamePtr := @thisVolumesName;
				thisVolumesName := '';
				err := PBHGetVInfoSync(@pb);
				if err <> noErr then begin
					leave;
				end; (* if *)
				if EqualString(volumeName, thisVolumesName, false, true) then begin
					if (pass = kNameOnlyMatchPass) | (pb.ioVCrDate = volumeCreationDate) then begin
						leave;
					end; (* if *)
				end; (* if *)
			end; (* while *)
			if err = noErr then begin
				leave;
			end; (* if *)
		end; (* for *)
		if err = noErr then begin
			vRefNum := pb.ioVRefNum;
		end; (* if *)
		FindVolumeByNameAndDate := err;
	end; (* FindVolumeByNameAndDate *)

	function ResolveICFileSpec(fileSpec : Handle; var resolvedFileSpec : FSSpec) : Boolean;
		(* This function attempts to resolve the fileSpec parameter, and
			points resolvedFileSpec the file it points to. The function returns
			true if it succeeds (ie resolvedFileSpec is valid), false otherwise.
		*)
		var
			result : Boolean;
			junkLong: longint;
			junkBool : Boolean;
			cpb: CInfoPBRec;
	begin
		result := true;			(* Ever optimisic! *)
		if (ICFileSpecHandle(fileSpec)^^.alias.aliasSize = 0) then begin
			(* There is no useful alias information in the fileSpec, let's
				do it the hard way. First find the volume, then assume the 
				name and dirID are OK, finally try to confirm the existence
				of the folder.
			*)
			if FindVolumeByNameAndDate(ICFileSpecHandle(fileSpec)^^.vol_name, 
															ICFileSpecHandle(fileSpec)^^.vol_creation_date,
															resolvedFileSpec.vRefNum) = noErr then begin
				resolvedFileSpec.parID := ICFileSpecHandle(fileSpec)^^.fss.parID;
				resolvedFileSpec.name := ICFileSpecHandle(fileSpec)^^.fss.name;
				if FSpGetCatInfo(resolvedFileSpec, 0, cpb) <> noErr then begin
					result := false;
				end; (* if *)
			end else begin
				result := false;
			end; (* if *)
		end else begin
		
			(* We have alias information, lets go resolve the alias. First remove
				the Poor Man's Alias info off the front of the ICFileSpecHandle, 
				then do the resolve.
			*)
			junkLong := Munger(fileSpec, 0, nil, sizeof(ICFileSpec) - sizeof(AliasRecord), @junkLong, 0);
			if ResolveAlias(nil, AliasHandle(fileSpec), resolvedFileSpec, junkBool) <> noErr then begin
				result := false;
			end; (* if *)
		end; (* if *)
		
		ResolveICFileSpec := result;
	end; (* ResolveICFileSpec *)

	procedure FindDesktopFolder(var desktopFolder : FSSpec);
		(* Put on your file system hat before reading this routine!
			It basically returns an FSSpec that points to the desktop
			folder, but you need to understand GetCatInfo to understand
			it.
		*)
		var
			err : OSStatus;
			cpb: CInfoPBRec;
	begin
		err := FindFolder(kOnSystemDisk, kDesktopFolderType, true, 
								desktopFolder.vRefNum, desktopFolder.parID);
		if err = noErr then begin
			(* Passing -1 to GetCatInfo causes it to only use
				the vRefNum and dirID fields (as setup above)
				to find the directory, and returns its parID
				and name.
			*)
			err := FSpGetCatInfo(desktopFolder, -1, cpb);
		end; (* if *)
		if err = noErr then begin
			desktopFolder.vRefNum := cpb.ioVRefNum;
			desktopFolder.parID := cpb.ioDrParID;
		end else begin
			desktopFolder.vRefNum := 0;
			desktopFolder.parID := 0;
			desktopFolder.name := '';
		end; (* if *)
	end; (* FindDesktopFolder *)

	(* ***** Standard What Interface ***** *)

	(* ***** What Open ***** *)

	function WhatOpenFSSpec (wt: WindowType; item: integer): OSStatus;
		(* See the comments in ICWindowGlobals. *)
		var
			fileSpec: Handle;		(* an ICFileSpecHandle in disguise *)
			fileSpecValid: Boolean;
			resolvedFileSpec: FSSpec;
			attr: longint;
	begin
		fileSpec := nil;
		
		(* Get the IC file fileSpec. *)
		fileSpecValid := false;
		if CacheGetPrefHandle(gWindows[wt].items[item]^.whatKey,
											attr, fileSpec) = noErr then begin
			fileSpecValid := true;
		end; (* if *)
		WhatProcessAttributes(wt, item, attr);

		(* Validate whether it's of a suitable size. *)
		if fileSpecValid & (GetHandleSize(fileSpec) < sizeof(fileSpec)) then begin
			fileSpecValid := false;
		end; (* if *)
		
		(* Now attempt to resolve it if it's fileSpecValid. *)
		if fileSpecValid then begin
			fileSpecValid := ResolveICFileSpec(fileSpec, resolvedFileSpec);
		end; (* if *)
		
		(* If it's still not fileSpecValid, point resolvedFileSpec to the desktop. *)
		if not fileSpecValid then begin
			FindDesktopFolder(resolvedFileSpec);
		end; (* if *)
		
		(* Finish up. *)
		if fileSpec <> nil then begin
			DisposeHandle(fileSpec);
		end; (* if *)
		gWindows[wt].items[item]^.whatFileSpec := resolvedFileSpec;
		SetDControlTitle(gWindows[wt].window, item, resolvedFileSpec.name);
		SetDControlEnable(gWindows[wt].window, item, not WhatIsLocked(wt, item));
		
		WhatOpenFSSpec := noErr;
	end; (* WhatOpenFSSpec *)

	(* ***** What Flush ***** *)

	function WhatFlushFSSpec (wt: WindowType; item: integer): OSStatus;
		(* See the comments in ICWindowGlobals. *)
		var
			err: OSStatus;
			icFileSpecPrefix: ICFileSpec;
			aliasH: AliasHandle;
			junkLong: longint;
			tmpStr : Str63;
	begin
		err := noErr;
		if not WhatIsLocked(wt, item) then begin
			(* Start filling out fields in the ICFileSpec. First the actual
				file spec component
			*)
			icFileSpecPrefix.fss := gWindows[wt].items[item]^.whatFileSpec;

			(* Now create the alias that's to go at the end of the
				final ICFileSpecHandle.
			*)
			aliasH := nil;
			err := NewAlias(nil, icFileSpecPrefix.fss, aliasH);
			if err <> noErr then begin
				(* Couldn't create the alias, instead create a
					blank pseudo-alias.
				*)
				aliasH := AliasHandle(NewHandle(sizeof(AliasRecord)));
				err := MemError;
				if err = noErr then begin
					aliasH^^.aliasSize := 0;
					aliasH^^.userType := OSType(0);
				end; (* if *)
			end; (* if *)
	
			(* OK, let's continue filling in the fields of the ICFileSpec.
				The final field is the volume creation date
			*)
			if err = noErr then begin

				// Use a temporary string because otherwise we have to pass
				// icFileSpecPrefix.vol_name (a Str31) to GetVolInfo (which
				// expects a Str63).  In practise, this isn't an issue, but
				// the compiler issues a warning (which is annoying, but correct
				// of it) and it could cause problems in the future, so we
				// might as well do The Right Thing (tm).
				
				tmpStr := '';
				err := GetVolInfo (tmpStr, icFileSpecPrefix.fss.vRefNum,
											0, icFileSpecPrefix.vol_creation_date);
				icFileSpecPrefix.vol_name := tmpStr;
			end; (* if *)
	
			(* All is cool, now insert the ICFileSpec at the front of the alias
				(or pseudo-alias, as the case may be).
			*)
			if err = noErr then begin
				junkLong := Munger(Handle(aliasH), 0, nil, 0, @icFileSpecPrefix, sizeof(ICFileSpec) - sizeof(AliasRecord));
				err := MemError;
			end; (* if *)
	
			(* And then save the preference and tidy up. *)
			if err = noErr then begin
				err := CacheSetPrefHandle(gWindows[wt].items[item]^.whatKey, Handle(aliasH));
			end; (* if *)
			if aliasH <> nil then begin
				DisposeHandle(Handle(aliasH));
			end; (* if *)
		end; (* if *)
		WhatFlushFSSpec := err;
	end; (* WhatFlushFSSpec *)

	(* ***** What Click ***** *)

	function WhatClickFSSpec (wt: WindowType; item: integer; event: EventRecord): OSStatus;
		(* See the comments in ICWindowGlobals.

			The big weirdness in this routine is that IC 1.3 and earlier
			put a deviant FSSpec into the whatFileSpec (namely a
			vRefNum/DirID/'' tuple rather than a vRefNum/ParID/DirName
			tuple).  This only affects System 6 users, because System
			7 users save and restore this setting via the alias, which
			isn't affected by the difference.
			
			I've tentatively changed this behaviour for newer versions
			of IC, but I've left the old code here in case problems arise
			in integration testing.
		*)
		var
			err: OSStatus;
			chosenFileSpec: FSSpec;
			dirIDOfChosenFileSpec: longint;
	begin
		{$unused event}
		err := ICStandardGetFolder(chosenFileSpec, dirIDOfChosenFileSpec);
		if err = noErr then begin
			{$ifc false}
				(* Old, incorrect behaviour.  I've left this around so that if
					the new behaviour breaks anybody, I'll be able to switch back
					quickly.  Note that System 7 clients use the alias info anyway,
					so there's little chance of breakage.
				*)
				gWindows[wt].items[item]^.whatFileSpec.vRefNum := chosenFileSpec.vRefNum;
				gWindows[wt].items[item]^.whatFileSpec.parID := dirIDOfChosenFileSpec;
				gWindows[wt].items[item]^.whatFileSpec.name := '';
			{$elsec}
				(* New, correct behaviour. *)
				gWindows[wt].items[item]^.whatFileSpec := chosenFileSpec;
			{$endc}
			SetDControlTitle(gWindows[wt].window, item, chosenFileSpec.name);
			WhatDirty(wt, item);
		end; (* if *)
		WhatClickFSSpec := err;
	end; (* WhatClickFSSpec *)

end. (* ICFSSpecWhats *)