unit ICIconSuites;

(*	This file is part of the Internet Configuration system and is placed in the public domain for the benefit of all.

	This module implements a desktop database (DTDB) icon cache.  Once you have
	initialised  the cache, you can ask for icon suites by their file type and creater.  
	The module looks up the icon in the DTDB and returns an icon suite, 
	which it also caches.  If the module fails to find a specific match in the DTDB,
	it returns either the generic document or the generic application icon.
	
	The module caches up to 20 icons, along with the time they were last accessed.
	Note that this time is not actually a real time; time is incremented each time
	the module is called.  When the cache is exhausted, suites are discarded in a least 
	recently used fashion.
	
	This module returns suites with only big icon information.  The extensions
	for small icons are left as an exercise to the user.
*)

interface

	uses
		Types,
		Icons;
		
	procedure InitICIconSuites;
		(* This routine initialises the module. *)
	
	procedure PlotIconCached(fileType : OSType; fileCreator: OSType;
							theRect : Rect; align : IconAlignmentType; 
							transform: IconTransformType;
							iconRegion : RgnHandle);
		(*	This routine plots an icon for a file of the given type
			and creator.  The Rect, alignment and transform are
			all equivalent to the similar parameters to PlotIconSuite.
			The iconRegion is set up to be the region that was plotted.
		*)

implementation

	uses
		Files,
		Errors, 
		Icons,
		Folders,

		ICDebug,		
		ICGlobals,
		ICMiscSubs,
		ICCommonSubs;
		
	(* ***** Internal Routines ***** *)

	function GetDTDBIcon (fileType, fileCreator: OSType; var cookie: integer; dtdbIconType: SignedByte;
												var dtpb: DTPBRec): OSStatus;
		(* This routine queries the DTDB for an icon of type dtdbIcon matching
			the fileType and fileCreator.  If it finds one, it returns it in the
			buffer specified in the dtpb.
		*)

		function GetTheIcon (dtrn: integer): Boolean;
			(* Attempts to get an icon of dtdbIconType from the DTDB file
				denoted by dtrn.
			*)
		begin
			dtpb.ioDTRefNum := dtrn;
			dtpb.ioTagInfo := 0;
			dtpb.ioIconType := dtdbIconType;
			dtpb.ioFileCreator := fileCreator;
			dtpb.ioFileType := fileType;
			GetTheIcon := PBDTGetIconSync(@dtpb) = noErr;
		end; (* GetTheIcon *)

		var
			err: OSStatus;
			found: Boolean;
			dtdbFileIndex: integer;
			vRefNum: integer;
			junkVolumeName: Str63;
			junkCreationDate: longint;
	begin
		found := false;
		
		(* If the cookie is empty, we don't know what DTDB file we're looking in,
			so we must go searching through the each DTDB files for the requested
			icon type.
		*)
		if cookie = 0 then begin
			dtdbFileIndex := 1;
			repeat
				(* Get the vRefNum associated with the dtdbFileIndex'th volume. *)
				vRefNum := 0;
				junkVolumeName := '';
				err := GetVolInfo(junkVolumeName, vRefNum, dtdbFileIndex, junkCreationDate);
				dtdbFileIndex := dtdbFileIndex + 1;
				if err = noErr then begin
				
					(* We've got a volume, now find the DTDB access path for it. *)
					dtpb.ioNamePtr := nil;
					dtpb.ioVRefNum := vRefNum;
					err := PBDTGetPath(@dtpb);
					if err = noErr then begin
					
						(* We have an access path, now try to get the icon. *)
						if GetTheIcon(dtpb.ioDTRefNum) then begin
							(* We got the icon, remember the access path in cookie and leave
								this mess.
							*)
							cookie := dtpb.ioDTRefNum;
							found := true;
						end; (* if *)
					end; (* if *)
					err := noErr;
				end; (* if *)
			until found or (err <> noErr);
		end else begin
			(* The cookie was not empty, therefore we know exactly which DTDB
				file to look in.  Go get the icon.
			*)
			found := GetTheIcon(cookie);
		end; (* if *)
		
		(* Clean up. *)
		if found then begin
			err := noErr;
		end else begin
			err := afpItemNotFound;
		end; (* if *)
		GetDTDBIcon := err;
	end; (* GetDTDBIcon *)

	var
		gTemporaryIconBuffer: packed array[0..1023] of Byte;

	function GetDTDBIconH (fileType, fileCreator: OSType; var cookie: integer; dtdbIconType: SignedByte; var iconH: Handle): OSStatus;
		(* This routine queries the DTDB for an icon of type dtdbIcon matching
			the fileType and fileCreator.  If it finds one, it returns it in a handle.
		*)
		var
			err: OSStatus;
			dtpb: DTPBRec;
	begin
		(* Get the icon into a global temporary icon buffer. *)
		iconH := nil;
		dtpb.ioDTBuffer := @gTemporaryIconBuffer;
		dtpb.ioDTReqCount := sizeof(gTemporaryIconBuffer);
		err := GetDTDBIcon(fileType, fileCreator, cookie, dtdbIconType, dtpb);
		
		(* If that succeeds, copy the icon buffer into a handle. *)
		if err = noErr then begin
			err := PtrToHand(@gTemporaryIconBuffer, iconH, dtpb.ioDTActCount);
		end; (* if *)
		
		(* Clean up. *)
		if err <> noErr then begin
			DisposeHandle(iconH);
			iconH := nil;
		end; (* if *)
		GetDTDBIconH := err;
	end; (* GetDTDBIconH *)

	function GetDTDBAddSuite (iconSuite: Handle; fileType, fileCreator: OSType; var cookie: integer;
											dtdbIconType: SignedByte; suiteIconType: OSType): OSStatus;
		(* This routine looks for the dtdbIconType icon corresponding to fileType
			and fileCreator and adds it into iconSuite.  Note that the DTDB and the
			icon suites use different codes to describe the various types of
			icons, which is why you have to pass two parameters in here.
		*)
		var
			err: OSStatus;
			iconH: Handle;
	begin
		err := GetDTDBIconH(fileType, fileCreator, cookie, dtdbIconType, iconH);
		if err = noErr then begin
			err := AddIconToSuite(iconH, iconSuite, suiteIconType);
		end; (* if *)
		if err <> noErr then begin
			DisposeHandle(iconH);
			iconH := nil;
		end; (* if *)
		GetDTDBAddSuite := err;
	end; (* GetDTDBAddSuite *)

	function GetDTDBIconSuiteUncached (fileType, fileCreator: OSType; var iconSuite: Handle): OSStatus;
		(* This routine looks in the DTDB for icons that correspond to fileType and
			fileCreator and returns them in an icon suite.  If to fails to find any
			icons it returns an error and iconSuite set to nil.
		*)
		var
			err: OSStatus;
			junk: OSStatus;
			cookie: integer;
	begin
		(* Create the icon suite. *)
		iconSuite := nil;
		err := NewIconSuite(iconSuite);
		
		(* Add icons from the DTDB into the suite.  Start with the 1 bit large icon.
			if that succeeds, then continue adding the colour large icons while
			junking the error codes.  If the large icon fails, stop immediately.
		*)
		if err = noErr then begin
			(* Cookie is passed to each of these routines as storage for those routines.
				The idea is that the DTDB routines can use this cookie to speed
				access to the appropriate DTDB file between calls.  Only the bottom
				level of the call chain knows are cares what cookie actually represents,
				namely the DTDB access path.
			*)
			cookie := 0;
			if GetDTDBAddSuite(iconSuite, fileType, fileCreator, cookie, kLargeIcon, large1BitMask) = noErr then begin
				junk := GetDTDBAddSuite(iconSuite, fileType, fileCreator, cookie, kLarge4BitIcon, large4BitData);
				junk := GetDTDBAddSuite(iconSuite, fileType, fileCreator, cookie, kLarge8BitIcon, large8BitData);
			end else begin
				err := afpItemNotFound;
			end; (* if *)
		end; (* if *)
		
		(* Clean up after failure. *)
		if err <> noErr then begin
			if iconSuite <> nil then begin
				junk := DisposeIconSuite(iconSuite, true);
				iconSuite := nil;
			end; (* if *)
		end; (* if *)
		GetDTDBIconSuiteUncached := err;
	end; (* GetDTDBIconSuiteUncached *)

	(* ***** Cache Data Structure ***** *)

	const
		kMaxCachedIcons = 20;

	type
		CacheRecord =
			record
				timeLastUsed: longint;
				fileType, fileCreator: OSType;
				iconSuite: IconSuiteRef;
				iconReference: IconRef;
			end;

	var
		gCachedIcons: array[1..kMaxCachedIcons] of CacheRecord;
		gCurrentTime: longint;
		gDefaultApplicationSuite: Handle;
		gDefaultDocumentSuite: Handle;

	(* ***** API Routines ***** *)

	procedure PlotIconCached(fileType : OSType; fileCreator: OSType;
							theRect : Rect; align : IconAlignmentType; 
							transform: IconTransformType;
							iconRegion : RgnHandle);
		(* This routine returns an icon suite handle for the given file type and creator.
			The icon suite is cached, so the caller does not need to worry about disposing
			it.  The suite is valid until the next time this module is called.

			If there is no explicit match in the DTDB for this icon, this routine sets
			iconSuite to the default document or application icon suite, but also 
			returns an error.
		*)
		var
			junk: OSStatus;
			iconIndex : integer;
			foundIconIndex : integer;
			oldestVictimTime : longint;
			tmpIconSuite : IconSuiteRef;
			tmpIconRef : IconRef;
	begin
		(* Start off by assuming failure. *)

		foundIconIndex := 0;
		
		(* Search the cache for the icon. *)
		for iconIndex := 1 to kMaxCachedIcons do begin
			if (gCachedIcons[iconIndex].timeLastUsed > 0) & (gCachedIcons[iconIndex].fileType = fileType) & (gCachedIcons[iconIndex].fileCreator = fileCreator) then begin
				foundIconIndex := iconIndex;
				gCachedIcons[iconIndex].timeLastUsed := gCurrentTime;
				gCurrentTime := gCurrentTime + 1;
				leave;
			end; (* if *)
		end; (* for *)
		
		(* If we didn't find it in the cache... *)
		
		if foundIconIndex = 0 then begin
			(*	Look for the least recently used icon in the cache.  This is the
				icon we intend to replace with the new icon were bringing in.
			*)
			oldestVictimTime := maxLongInt;
			for iconIndex := 1 to kMaxCachedIcons do begin
				if (gCachedIcons[iconIndex].timeLastUsed < oldestVictimTime) then begin
					foundIconIndex := iconIndex;
					oldestVictimTime := gCachedIcons[iconIndex].timeLastUsed;
				end; (* if *)
			end; (* for *)
			
			(* 	We must have found a victim because oldestVictimTime started as
				maxlongint, meaning that all icons are potential victims.
			*)
			ICAssert(foundIconIndex <> 0);
			
			(*	We implement the same algorithm for both Icon Utilities and
				Icon Services.  The idea is to get the icon from the API
				and, if that succeeds, dispose of the existing icon in the
				cache and replace it with the new one, leaving foundIconIndex
				pointing to the entry we just put the new icon in.  If we
				cant get an icon, leave the cache entry alone and set foundIconIndex
				to 0 to indicate that.
			*)
			
			if gHasIconServices then begin
				ICAssert(gCachedIcons[foundIconIndex].iconSuite = nil);			// Were using Icon Services, there should be no icon suites in the cache.
				if GetIconRef(kOnSystemDisk, fileCreator, fileType, tmpIconRef) = noErr then begin
					if gCachedIcons[foundIconIndex].iconReference <> nil then begin
						junk := ReleaseIconRef(gCachedIcons[foundIconIndex].iconReference);
						ICAssert(junk = noErr);
					end; (* if *)
					gCachedIcons[foundIconIndex].iconReference := tmpIconRef;
				end else begin
					foundIconIndex := 0;
				end; (* if *)
			end else begin
				ICAssert(gCachedIcons[foundIconIndex].iconReference = nil);		// Were using Icon Utilities, there should be no icon refs in the cache.
				if GetDTDBIconSuiteUncached(fileType, fileCreator, tmpIconSuite) = noErr then begin
					if gCachedIcons[foundIconIndex].iconSuite <> nil then begin
						junk := DisposeIconSuite(gCachedIcons[foundIconIndex].iconSuite, true);
						ICAssert(junk = noErr);
					end; (* if *)
					gCachedIcons[foundIconIndex].iconSuite := tmpIconSuite;
				end else begin
					foundIconIndex := 0;
				end; (* if *)
			end; (* if *)
			
			// Fill out the common entries in the cache record.
			
			if foundIconIndex <> 0 then begin
				gCachedIcons[foundIconIndex].fileType := fileType;
				gCachedIcons[foundIconIndex].fileCreator := fileCreator;
				gCachedIcons[foundIconIndex].timeLastUsed := gCurrentTime;
				gCurrentTime := gCurrentTime + 1;
			end; (* if *)
			
		end; (* if *)
		
		(*	Now plot the icon out of the cache entry, if we could find/create
			one.  If not, plot some defaults.
		*)
		
		if foundIconIndex <> 0 then begin
			if gHasIconServices then begin
				junk := PlotIconRef(theRect, align, transform, kIconServicesNormalUsageFlag, gCachedIcons[foundIconIndex].iconReference);
				ICAssert(junk = noErr);
				junk := IconRefToRgn(iconRegion, theRect, align, kIconServicesNormalUsageFlag, gCachedIcons[foundIconIndex].iconReference);
				ICAssert(junk = noErr);
			end else begin
				junk := PlotIconSuite(theRect, align, transform, gCachedIcons[foundIconIndex].iconSuite);
				ICAssert(junk = noErr);
				junk := IconSuiteToRgn(iconRegion, theRect, align, gCachedIcons[foundIconIndex].iconSuite);
				ICAssert(junk = noErr);
			end; (* if *)
		end else begin
			tmpIconSuite := gDefaultDocumentSuite;
			if IsApplicationType(fileType) & (gDefaultApplicationSuite <> nil) then begin
				tmpIconSuite := gDefaultApplicationSuite;
			end; (* if *)
			junk := PlotIconSuite(theRect, align, transform, tmpIconSuite);
			ICAssert(junk = noErr);
			junk := IconSuiteToRgn(iconRegion, theRect, align, tmpIconSuite);
			ICAssert(junk = noErr);
		end; (* if *)
	end; (* PlotIconCached *)

	procedure InitICIconSuites;
		var
			cacheIndex: integer;
	begin
		gCurrentTime := 1;

		(* Clear out all of the entries in the cache. *)
		for cacheIndex := 1 to kMaxCachedIcons do begin
			gCachedIcons[cacheIndex].timeLastUsed := -1;
			gCachedIcons[cacheIndex].iconSuite := nil;
			gCachedIcons[cacheIndex].iconReference := nil;
		end; (* if *)
		
		(* Setup the default application and default document suites. The icon
			data for these icons comes from the system file.
		*)
		gDefaultApplicationSuite := nil;
		gDefaultDocumentSuite := nil;
		if GetIconSuite(gDefaultDocumentSuite, -4000, svAllLargeData) <> noErr then begin
			gDefaultDocumentSuite := nil;
		end; (* if *)
		if GetIconSuite(gDefaultApplicationSuite, -3996, svAllLargeData) <> noErr then begin
			gDefaultApplicationSuite := nil;
		end; (* if *)
	end; (* InitICIconSuites *)

end. (* ICIconSuites *)