unit ICPreferenceCache;

(*	This file is part of the Internet Configuration system and is placed in the public domain for the benefit of all.

	This file is used to cache modified preferences.  When IC accesses preferences,
	it always does it through this module.  When IC modifies a preference, this module
	stores the modified preference in memory, only writing it to the actual IC
	preferences file when you commit.
	
	As the client application, ie Internet Config, only ever writes preferences without
	changing the attributes, we assume ICattr_no_change in all cases, which makes
	the implementation a bit easier.
	
	This module is staggeringly inefficient in both memory and speed.  It burns
	memory by storing a Str255 for the preference key, rather than storing it
	in some packed fashion.  It burns speed by storing all modified preferences
	in a single unsorted array.  It was, however, very easy to write (-:
*)
 
interface

	uses
		Types,

		InternetConfig;
		
	function  CacheGetPref(const key : Str255; var attr : ICAttr; buf : Ptr; var size : longint) : OSStatus;
	function  CacheSetPref(const key : Str255;                    buf : Ptr; size : longint) : OSStatus;

	function  CacheGetPrefStr(const key : Str255; var attr : ICAttr;   var str : Str255) : OSStatus;
	function  CacheSetPrefStr(const key : Str255;                    const str : Str255) : OSStatus;

	function  CacheGetPrefHandle(const key : Str255; var attr : ICAttr; var prefH : Handle) : OSStatus;
	function  CacheSetPrefHandle(const key : Str255;                        prefH : Handle) : OSStatus;

	function  CacheGetIndPref(n : longint; var key : Str255) : OSStatus;
	
	function  CacheDeletePref(const key : Str255) : OSStatus;
	
	function  CacheSetCurrentProfile(newID : ICProfileID) : OSStatus;
	function  CacheGetCurrentProfile : ICProfileID;
	
	function  CacheCountProfiles : longint;
	
	function  CacheCommitChanges : OSStatus;
	function  CacheDiscardChanges : OSStatus;
	
	function  InitCache(inst : ICInstance) : OSStatus;
	procedure TermCache;

implementation

	uses
		Memory,
		StringCompare,
		
		ICDebug,
		ICCommonSubs;

	type
		CacheIndex0 = 0..5000;
		CacheIndex = 1..5000;
		CacheEntry =
			record
				key : Str255;
				attr : ICAttr;
				data : Handle;
				deleted : Boolean;
				dirty : Boolean;
			end;
		CacheEntryArray = array [CacheIndex] of CacheEntry;
		CachePtr = ^CacheEntryArray;
		CacheHandle = ^CachePtr;
		
	var
		gCacheInstance : ICInstance;
		gCache : CacheHandle;
		gCurrentProfile : ICProfileID;
		gLastProfileCount : longint;

	function  CacheSize : longint;
	begin
		ICAssert(gCache <> nil);
		CacheSize := GetHandleSize(Handle(gCache)) div sizeof(CacheEntry);
	end; (* CacheSize *)
	
	function  FindCacheEntry(const key : Str255; findDeleted : Boolean) : CacheIndex0;
		var
			index : CacheIndex0;
			tmpKey : Str255;
			thisEntry : CacheEntry;
	begin
		tmpKey := key;
		FindCacheEntry := 0;
		for index := 1 to CacheSize do begin
			thisEntry := gCache^^[index];
			if EqualString(tmpKey, thisEntry.key, false, true) & (findDeleted | not thisEntry.deleted) then begin
				FindCacheEntry := index;
				leave;
			end; (* if *)
		end; (* for *)
	end; (* FindCacheEntry *)
		
	function  CacheGetPref(const key : Str255; var attr : ICAttr; buf : Ptr; var size : longint) : OSStatus;
		var
			err : OSStatus;
			foundIndex : CacheIndex0;
			maxSizeOfClientBuffer : longint;
	begin
		foundIndex := FindCacheEntry(key, false);
		if foundIndex = 0 then begin
			attr := ICattr_no_change;
			size := 0;
			err := icPrefNotFoundErr;
		end else begin
		
			// Its in our cache, lets copy the data out of the cache
			// rather than out of IC.  The only tricky part is ensuring
			// that we set up size correctly and dont write off the end
			// of the client buffer.
			
			err := noErr;

			maxSizeOfClientBuffer := size;
			
			attr := gCache^^[foundIndex].attr;
			size := GetHandleSize(gCache^^[foundIndex].data);
			if buf <> nil then begin
				if size > maxSizeOfClientBuffer then begin
					err := icTruncatedErr;
				end else begin
					maxSizeOfClientBuffer := size;
				end; (* if *)
				BlockMoveData(gCache^^[foundIndex].data^, buf, maxSizeOfClientBuffer);
			end; (* if *)
		end; (* if *)
		
		CacheGetPref := err;
	end; (* CacheGetPref *)

	function  CacheSetPref(const key : Str255;                    buf : Ptr; size : longint) : OSStatus;
		var
			err : OSStatus;
			foundIndex : CacheIndex0;
			thisEntry : CacheEntry;
	begin
		foundIndex := FindCacheEntry(key, true);
		if foundIndex = 0 then begin

			// Its not already in the cache.  Add an entry for it.

			// Fill out thisEntry, including a handle to the data,
			// and then add it to the cache.
			
			thisEntry.key := key;
			thisEntry.attr := 0;
			thisEntry.deleted := false;
			thisEntry.dirty := true;
			err := PtrToHand(buf, thisEntry.data, size);
			if err = noErr then begin
				err := PtrAndHand(@thisEntry, Handle(gCache), sizeof(thisEntry));
				if err <> noErr then begin
					ICAssert(thisEntry.data <> nil);
					DisposeHandle(thisEntry.data);
					ICAssert(MemError = noErr);
				end; (* if *)
			end; (* if *)
							
		end else begin
		
			// Its already in the cache, lets just update
			// the entry.  profile and key must be unchanged,
			// theres no way for the client to change attr,
			// so the only change is the client data.
			
			err := PtrToXHand(buf, gCache^^[foundIndex].data, size);
			
			// If the update succeeded, we now know the entry is
			// no longer deleted and is dirty.
			
			if err = noErr then begin
				gCache^^[foundIndex].deleted := false;
				gCache^^[foundIndex].dirty := true;
			end; (* if *)
		end; (* if *)
		
		CacheSetPref := err;
	end; (* CacheSetPref *)
	
{$PUSH}
{$R-}
	function  CacheGetPrefStr(const key : Str255; var attr : ICAttr;   var str : Str255) : OSStatus;
		var
			err : OSStatus;
			size : longint;
	begin
		size := 256;
		err := CacheGetPref(key, attr, @str, size);
		if err <> noErr then begin
			str := '';
		end; (* if *)
		CacheGetPrefStr := err;
	end; (* CacheGetPrefStr *)

	function  CacheSetPrefStr(const key : Str255;                    const str : Str255) : OSStatus;
	begin
		CacheSetPrefStr := CacheSetPref(key, @str, length(str) + 1);
	end; (* CacheSetPrefStr *)
{$POP}

	function  CacheGetPrefHandle(const key : Str255; var attr : ICAttr; var prefH : Handle) : OSStatus;
		var
			err : OSStatus;
			preferenceSize : longint;
	begin
		prefH := NewHandle(0);
		err := CheckMemError(prefH);
		if err = noErr then begin
			preferenceSize := 0;
			
			err := CacheGetPref(key, attr, nil, preferenceSize);
			if err = noErr then begin
				SetHandleSize(prefH, preferenceSize);
				err := MemError;
			end; (* if *)
			if err = noErr then begin
				HLock(prefH);
				err := CacheGetPref(key, attr, prefH^, preferenceSize);
				HUnlock(prefH);
			end; (* if *)

			if err <> noErr then begin
				SetHandleSize(prefH, 0);
			end; (* if *)
		end; (* if *)
		
		// If the preferences isnt found, return noErr and an empty handle.
		
		if err = icPrefNotFoundErr then begin
			SetHandleSize(prefH, 0);
			attr := 0;
			err := noErr;
		end; (* if *)
		CacheGetPrefHandle := err;
	end; (* CacheGetPrefHandle *)

	function  CacheSetPrefHandle(const key : Str255;                        prefH : Handle) : OSStatus;
		var
			err : OSStatus;
			s : SInt8;
	begin
		ICAssert(prefH <> nil);
		ICAssert(prefH^ <> nil);
		
		s := HGetState(prefH);
		HLock(prefH);
		err := CacheSetPref(key, prefH^, GetHandleSize(prefH));
		HSetState(prefH, s);

		CacheSetPrefHandle := err;
	end; (* CacheSetPrefHandle *)

	function  CacheGetIndPref(n : longint; var key : Str255) : OSStatus;
		var
			err : OSStatus;
			numPrefs : longint;
			thisPref : longint;
			indexOfThisPref : longint;
			found : Boolean;
	begin
		err := noErr;
		ICParamCheck(err, n >= 1, 'index must be positive');
		if err = noErr then begin
			numPrefs := CacheSize;
			thisPref := 1;
			indexOfThisPref := 1;
			found := false;
			while (thisPref <= numPrefs) & not found do begin
				if gCache^^[thisPref].deleted then begin
					thisPref := thisPref + 1;
				end else begin
					found := (indexOfThisPref = n);
					if not found then begin
						indexOfThisPref := indexOfThisPref + 1;
						thisPref := thisPref + 1;
					end; (* if *)
				end; (* if *)
			end; (* while *)
			
			if found then begin
				key := gCache^^[thisPref].key;
			end else begin
				err := icPrefNotFoundErr;
			end; (* if *)
		end; (* if *)
		CacheGetIndPref := err;
	end; (* CacheGetIndPref *)

	function  CacheDeletePref(const key : Str255) : OSStatus;
		var
			err : OSStatus;
			foundIndex : CacheIndex0;
	begin
		foundIndex := FindCacheEntry(key, false);
		if foundIndex = 0 then begin
			err := icPrefNotFoundErr;
		end else begin
			SetHandleSize(gCache^^[foundIndex].data, 0);
			ICAssert(MemError = noErr);
			gCache^^[foundIndex].deleted := true;
			err := noErr;
		end; (* if *)

		CacheDeletePref := err;
	end; (* CacheDeletePref *)
	
	procedure EmptyCache;
		// This routine empties the cache competely, discarding
		// any data that was in it.  It is called by various
		// routines before calling FillCacheWithProfile to remove
		// prior contents of the cache, and by the termination code.
		var
			index : CacheIndex0;
	begin
		for index := 1 to CacheSize do begin
			ICAssert(gCache^^[index].data <> nil);
			DisposeHandle(gCache^^[index].data);
			ICAssert(MemError = noErr);
		end; (* for *)
		SetHandleSize(Handle(gCache), 0);
		ICAssert(MemError = noErr);
	end; (* EmptyCache *)
	
	function FillCacheWithProfile : OSStatus;
		// This routine fills the cache with the preferences
		// in the profile gCurrentProfile.
		var
			err : OSStatus;
			err2 : OSStatus;
			oldProfile : ICProfileID;
			didSwitchProfiles : Boolean;
			currentPerm : ICPerm;
			didBegin : Boolean;
			prefCount : longint;
			prefIndex : longint;
			thisEntry : CacheEntry;
	begin
		ICAssert(CacheSize = 0);
		
		// First, preserve the old profile and switch to the one
		// were currently targetting (only if necessary).
		
		didSwitchProfiles := false;
		didBegin := false;
		err := ICGetCurrentProfile(gCacheInstance, oldProfile);
		if (err = noErr) & (oldProfile <> gCurrentProfile) then begin
			err := ICSetCurrentProfile(gCacheInstance, gCurrentProfile);
			didSwitchProfiles := (err = noErr);
		end; (* if *)

		// Now iterate through all the preferences, reading
		// each one into the cache.
		
		if err = noErr then begin
			if ( ICGetPerm(gCacheInstance, currentPerm) = noErr ) & (currentPerm = icNoPerm) then begin
				err := ICBegin(gCacheInstance, icReadOnlyPerm);
				didBegin := (err = noErr);
			end; (* if *)
		end; (* if *)
		if err = noErr then begin
			if ICCountProfiles(gCacheInstance, gLastProfileCount) <> noErr then begin
				gLastProfileCount := 1;
			end; (* if *)
		end; (* if *)
		if err = noErr then begin
			err := ICCountPref(gCacheInstance, prefCount);
		end; (* if *)
		if err = noErr then begin
			for prefIndex := 1 to prefCount do begin
			
				// Construct thisEntry to represent the entry
				// to add and then add it to the gCache handle.
				
				thisEntry.attr := 0;
				thisEntry.deleted := false;
				thisEntry.dirty := false;
				thisEntry.data := NewHandle(0);
				err := CheckMemError(thisEntry.data);
				if err = noErr then begin
					err := ICGetIndPref(gCacheInstance, prefIndex, thisEntry.key);
				end; (* if *)
				if err = noErr then begin
					err := ICFindPrefHandle(gCacheInstance, thisEntry.key, thisEntry.attr, thisEntry.data);
				end; (* if *)
				if err = noErr then begin
					err := PtrAndHand(@thisEntry, Handle(gCache), sizeof(thisEntry));
				end; (* if *)
				
				// Clean up on failure.
				
				if err <> noErr then begin
					if thisEntry.data <> nil then begin
						DisposeHandle(thisEntry.data);
						ICAssert(MemError = noErr);
					end; (* if *)
					leave;
				end; (* if *)
			end; (* for *)
		end; (* if *)
		
		if didBegin then begin
			err2 := ICEnd(gCacheInstance);
			if err = noErr then begin
				err := err2;
			end; (* if *)
		end; (* if *)
		
		// Switch back to the starting profile if necessary.
		
		if didSwitchProfiles then begin
			err2 := ICSetCurrentProfile(gCacheInstance, oldProfile);
			if err = noErr then begin
				err := err2;
			end; (* if *)
		end; (* if *)
		
		FillCacheWithProfile := err;
	end; (* FillCacheWithProfile *)

	function  CacheGetCurrentProfile : ICProfileID;
	begin
		CacheGetCurrentProfile := gCurrentProfile;
	end; (* CacheGetCurrentProfile *)

	function  CacheCountProfiles : longint;
	begin
		CacheCountProfiles := gLastProfileCount;
	end; (* CacheCountProfiles *)

	function  CacheSetCurrentProfile(newID : ICProfileID) : OSStatus;
		var
			err : OSStatus;
	begin
		EmptyCache;
		err := ICSetCurrentProfile(gCacheInstance, newID);
		if err = noErr then begin
			gCurrentProfile := newID;
			err := FillCacheWithProfile;
		end; (* if *)
		CacheSetCurrentProfile := err;
	end; (* CacheSetCurrentProfile *)
		
	function  CacheCommitChanges : OSStatus;
		var
			err : OSStatus;
			err2 : OSStatus;
			oldProfile : ICProfileID;
			didSwitchProfiles : Boolean;
			index : CacheIndex0;
			thisEntry : CacheEntry;
			s : SInt8;
	begin
		// First, preserve the old profile and switch to the one
		// were currently targetting (only if necessary).

		didSwitchProfiles := false;
		err := ICGetCurrentProfile(gCacheInstance, oldProfile);
		if (err = noErr) & (oldProfile <> gCurrentProfile) then begin
			err := ICSetCurrentProfile(gCacheInstance, gCurrentProfile);
			didSwitchProfiles := (err = noErr);
		end; (* if *)

		if err = noErr then begin
			err := ICBegin(gCacheInstance, icReadWritePerm);
		end; (* if *)
		if err = noErr then begin
		
			// Iterate over all the cached preferences, writing them to IC.
		
			for index := 1 to CacheSize do begin
				thisEntry := gCache^^[index];
				
				if thisEntry.deleted then begin
					err := ICDeletePref(gCacheInstance, thisEntry.key);
					if err = icPrefNotFoundErr then begin
						err := noErr;
					end; (* if *)
				end else if thisEntry.dirty then begin
					s := HGetState(thisEntry.data);
					HLock(thisEntry.data);
					err := ICSetPref(gCacheInstance, thisEntry.key, ICattr_no_change,
										thisEntry.data^, GetHandleSize(thisEntry.data));
					HSetState(thisEntry.data, s);
					
					gCache^^[index].dirty := (err = noErr);
				end; (* if *)
				
				if err <> noErr then begin
					leave;
				end; (* if *)
			end; (* for *)

			err2 := ICEnd(gCacheInstance);
			if err = noErr then begin
				err := err2;
			end; (* if *)
		end; (* if *)

		// Switch back to the starting profile if necessary.
		
		if didSwitchProfiles then begin
			err2 := ICSetCurrentProfile(gCacheInstance, oldProfile);
			if err = noErr then begin
				err := err2;
			end; (* if *)
		end; (* if *)

		CacheCommitChanges := err;
	end; (* CacheCommitChanges *)
	
	function CacheDiscardChanges : OSStatus;
		var
			err : OSStatus;
	begin
		EmptyCache;
		err := FillCacheWithProfile;
		CacheDiscardChanges := err;
	end; (* CacheDiscardChanges *)

	function  InitCache(inst : ICInstance) : OSStatus;
		var
			err : OSStatus;
	begin
		ICAssert(inst <> nil);
		
		gCacheInstance := inst;
		gCache := CacheHandle(NewHandle(0));
		err := CheckMemError(gCache);
		if err = noErr then begin
			err := ICGetCurrentProfile(gCacheInstance, gCurrentProfile);
		end; (* if *)
		if err = noErr then begin
			err := FillCacheWithProfile;
		end; (* if *)
		
		// Clean up.
		
		if err <> noErr then begin
			TermCache;
		end; (* if *)
		
		InitCache := err;
	end; (* InitCache *)
	
	procedure TermCache;
	begin
		if gCache <> nil then begin
			EmptyCache;
			
			DisposeHandle(Handle(gCache));
			ICAssert(MemError = noErr);
			gCache := nil;
		end; (* if *)
		gCacheInstance := nil;
	end; (* TermICCache *)

end. (* ICPreferenceCache *)