unit ICStandardFile;

(*	This file is part of the Internet Configuration system and is placed in the public domain for the benefit of all.

	This routine provides a set of wrappers around the system Standard File
	routines.  We need these because:
	
	1. We have to use Navigation Services if it's available, or Standard File
	   if it isn't.  This includes not linking to any Standard File routines
	   if we're compiling for Carbon.
	2. The system routine's API are less simple than we need.
	3. There's no system StandardGetFolder for Standard File.

	All of these routines return userCanceledErr if the user cancels the dialog.
	
	One nasty trick in this module is the Point($00640064) construct.  This
	creates a Point data type where each co-ordinate is 100 decimal, perfect
	for putting the standard file dialog in a place you can see.  We only use
	this for StandardGetFolder dialog.
*)

interface

	uses
		Files,
		
		ICResources,
		ICDialogs;
	
	function ICStandardGetFile (requestedFileType: OSType; 
												var chosenFileSpec: FSSpec;
												var infoForChosenFile: FInfo): OSStatus;
		(* This routine prompts the user for a file of type requestedFileType and returns
			a reference to that file in chosenFileSpec.  It also returns the file's Finder Info,
			because routines that call this routine often need that information.
		*)
	
	function ICStandardPutFile(prompt : Str255; name: Str255; fileType, fileCreator : OSType;
								var chosenFileSpec: FSSpec; var cookie : Ptr): OSStatus;
		(* This routine prompts the user for a destination for a file to be saved.
			The call provides a prompt string for the dialog and a default name string
			for the file.  The file's final destination is return in chosenFileSpec.
		*)
	
	function ICStandardCompleteSave(cookie : Ptr) : OSStatus;

	function ICStandardGetFolder (var chosenFolderSpec: FSSpec;
														var dirIDOfChosenFolder: longint): OSStatus;
		(* This routine prompts the user to choose a folder.  It returns
			a reference to the chosen folder.  It also returns the dirID of the chosen
			folder (as opposed to the parID in the FSSpec, which is the dirID of
			the folder that contains the chosen folder).  The jury is still up in the
			air as to whether this is needed by ICStandardGetFolder single
			client (the FSSpecWhat), but until I decide that issue, the parameter
			lives.
		*)

	function ICStandardAskSaveChanges(documentName : Str255; quitting : Boolean;
										var saveThem : Boolean) : OSStatus;
		(* A wrapper for the "Do you want to save changes?" dialog, which
			allows us to call NavAskSaveChanges on systems which have it.
		*)
		
	procedure InitICStandardFile;
		(* Initialises the module. *)

	procedure TermICStandardFile;
		(* Shut down the module. *)
		
implementation

	uses
		Finder,
		StandardFile, 
		LowMem,
		TextUtils,
		Aliases,
		Translation,
		Navigation,
		{$IFC NOT TARGET_API_MAC_CARBON} ICAccessors, {$ENDC}
		
		ICDebug,
		ICGlobals, 
		ICMiscSubs,
		ICCommonSubs,
		ICStringLists,
		ICDialogs,
		
		ICEvents;

{$ifc not TARGET_API_MAC_CARBON}

	// None of this stuff will even compile under Carbon, so
	// we have to conditionalise it out.

	var
		gStandardGetFolderFileFilterUPP : FileFilterYDUPP;
		gStandardGetFolderDialogUPP : DlgHookYDUPP;
		
	(* ***** StandardGetFolder for System 7 ****** *)

	(* Original comment:
	
		This little chunk o' code implements a way to let the user choose a       
		folder to save files in via a StandardFile Dialog.

		Since the code uses the CustomGetFile function and depends on the use of  
		FSSpec records, it only works under System 7.0 or later.                  
  
		Portions of this code were originally provided by Paul Forrester
		<paulf@apple.com> to the think-c Internet mailing list in response to my
		my question on how to do exactly what this code does.  I've added a
		couple of features, such as the ability to Handle aliased folders and
		the programmer definable prompt.  I also cleaned and tightened the code,
		stomped a couple of bugs, and packaged it up neatly.  Bunches of work,
		but I learned A LOT about Standard File, the File Manager, the Dialog
		Manager, and the Alias Manager.  I tried to include in the comments some
		of the neat stuff I discovered in my hours of pouring over Inside Mac.
		Hope you find it educational as well as useful.

		Converted to Pascal by Peter N Lewis <peter.lewis@info.curtin.edu.au>, Dec 1992.
		
		Cleaned up and integrated into IC by Quinn <quinn@quinn.echidna.id.au>,  Oct 1996
	*)

	type
		StandardFileReplyPtr = ^StandardFileReply;

	var
		gCurrentSelectedFolder: Str255;
		gStandardGetFolderPrompt: Str255;

	function StandardGetFolderFileFilter (cpbPtr: CInfoPBPtr; ignored: Ptr): Boolean;
		(* This is the file filter passed to CustomGetFile. It passes folders only. *)
	begin
		{$unused ignored}
		StandardGetFolderFileFilter := (band(cpbPtr^.ioFlAttrib, ioDirMask) = 0) | 
															(band(cpbPtr^.ioDrUsrWds.frFlags, $4000) <> 0);
	end; (* StandardGetFolderFileFilter *)

	function StandardGetFolderDialogHook (hitItem: integer; dlg: DialogPtr; myDataPtr: StandardFileReplyPtr) : integer;
		(* This function lets us process item hits in the GetFolderDialog.  We're
			only interested if the user hit the ditGetFolderSelectButton. We pass all
			other item hits back unmodified.
		*)

		procedure SetButtonTitle (newName: Str255);
			(* This function sets the title of the "Select Wibble" button to newName
				truncated appropriately.  It also updates the gCurrentSelectedFolder
				with the full name, so that we can check whether we need to
				do this work the next time around.
			*)
			var
				buttonNameProto : Str255;
				placeHolderIndex: integer;
				leftPart, rightPart: Str255;
				itemRect: Rect;
				junkResultCode: integer;
				widthRemaining: integer;
				junk : OSStatus;
		begin
			if gCurrentSelectedFolder <> newName then begin
				gCurrentSelectedFolder := newName;

				(* Get the prototype for the button name.  This is stored in the
					text of an offscreen static text item in the dialog. Then
					pull it apart by searching for ^1 and copying out the left and
					right parts.
				*)
				GetDItemText(dlg, ditGetFolderSelectString, buttonNameProto);
				
				placeHolderIndex := pos('^1', buttonNameProto);
				leftPart := TPCopy(buttonNameProto, 1, placeHolderIndex - 1);
				rightPart := TPCopy(buttonNameProto, placeHolderIndex + 2, 255);
				
			    (* Find the width left over in the button after drawing the word "Select"
			    	the quotation marks. Truncate the new name to this width.
			    *)
				GetDItemRect(dlg, ditGetFolderSelectButton, itemRect);
				widthRemaining := (itemRect.right - itemRect.left) - StringWidth(concat(' ', leftPart, rightPart, ' '));
				junkResultCode := TruncString(widthRemaining, newName, smTruncEnd);

				(* Finally set the buttons title. *)
				SetDControlTitle(dlg, ditGetFolderSelectButton, concat(leftPart, newName, rightPart));
				junk := ValidWindowRect(GetDialogWindow(dlg), itemRect);
				ICAssert(junk = noErr);
			end; (* if *)
		end; (* SetButtonTitle *)

		procedure SetFolderButtonTitle (vRefNum: integer; dirID: longint);
			(* This function sets the name of the "Select wibble" button
				based on the vRefNum and and dirID of a folder.
			*)
			var
				err: OSStatus;
				cpb: CInfoPBRec;
				folderName: Str63;
		begin
			cpb.ioFDirIndex := -1;
			cpb.ioNamePtr := @folderName;
			cpb.ioVRefNum := vRefNum;
			cpb.ioDirID := dirID;
			err := PBGetCatInfoSync(@cpb);
			if err = noErr then begin
				SetButtonTitle(folderName);
			end; (* if *)
		end; (* SetFolderButtonTitle *)

	begin
	    (* CustomGetFile calls dialog hook for both main and subsidiary dialog boxes.
			Make sure that dialog record indicates that this is the main GetFolder dialog.
		*)
		if OSType(GetWRefCon(GetDialogWindow(dlg))) = sfMainDialogRefCon then begin

			if hitItem = sfHookFirstCall then begin
				(* This is the first call, so setup the prompt above the file list. *)
				SetDItemText(dlg, ditGetFolderPromptString, gStandardGetFolderPrompt);
			end else begin
				(* This is a subsequent call. *)
				if myDataPtr^.sfFile.name = '' then begin
					(* If the current name is empty, the following fields aren't always set properly
						so get them out of low memory.
					*)
					myDataPtr^.sfFile.vRefNum := - LMGetSFSaveDisk;
					myDataPtr^.sfFile.parID := LMGetCurDirStore;
					myDataPtr^.sfFile.name := '';
					SetFolderButtonTitle(myDataPtr^.sfFile.vRefNum, myDataPtr^.sfFile.parID);
				end else begin
					SetButtonTitle(myDataPtr^.sfFile.name);
				end; (* if *)
			end; (* if *)

			(* If they click on the "Select Wibble" button, tell Standard File to bail
				out by returning the Cancel button, but set the sfGood Boolean so
				that we know to use the results.
			*)
			if hitItem = ditGetFolderSelectButton then begin
				hitItem := sfItemCancelButton;
				myDataPtr^.sfGood := true;
			end; (* if *)

		end; (* if *)
		StandardGetFolderDialogHook := hitItem;
	end; (* StandardGetFolderDialogHook *)

	procedure StandardGetFolder (prompt : Str255; var mySFReply: StandardFileReply);
		(* The StandardGetFolder function. You pass it the prompt to display above the file
			list, and a place to put a StandardFileReply record.

			Upon return, the sfFile field of the SFReply record contains the file
			spec of the chosen folder.  The sfGood field is set to true if the user
			chose a folder, or false if not.
		*)
		var
			err: OSStatus;
			dummyTypeList: SFTypeList;
			junkBool1, junkBool2: Boolean;
			cpb: CInfoPBRec;
	begin
		gStandardGetFolderPrompt := prompt;
		gCurrentSelectedFolder := '';
		CustomGetFile(gStandardGetFolderFileFilterUPP,	(* filters out everything except folders *)
						-1, @dummyTypeList, 						(* dummy type list *)
						mySFReply, 										(* space for result *)
						rCustomGetFolderDialog, 					(* dialog resource *)
						Point($00640064),							(* place to put window, system overrides this *)
						gStandardGetFolderDialogUPP, 		(* dialog hook drives our bonus items *)
						nil, 													(* no modal filter proc *)
						nil, 													(* no activation order list *)
						nil, 													(* no activate proc *)
						@mySFReply);									(* our data pointer *)

	    (* Now the reply record contains the volume reference number and the
			name of the selected folder. We need to use PBGetCatInfo to get the
			directory ID of the selected folder.
		*)
		if mySFReply.sfGood then begin
			
			(* If we got a name-based response (ie the user clicked the Open button
				rather than the Select Wibble button), make sure to resolve any aliases
				before continuing.
			*)
			if mySFReply.sfFile.name <> '' then begin
				err := ResolveAliasFile(mySFReply.sfFile, true, junkBool1, junkBool2);
				if err = noErr then begin
					(* Now look up the dirID of the folder we're pointed at. *)
					cpb.ioVRefNum := mySFReply.sfFile.vRefNum;
					cpb.ioDirID := mySFReply.sfFile.parID;
					cpb.ioNamePtr := @mySFReply.sfFile.name;
					cpb.ioFDirIndex := 0;
					err := PBGetCatInfoSync(@cpb);
				end; (* if *)

				mySFReply.sfGood := (err = noErr);
				mySFReply.sfFile.parID := cpb.ioDrDirID;
				mySFReply.sfFile.name := '';
			end; (* if *)
			
			(* Now call GetCatInfo to find the real FSSpec for the vRefNum/dirID pair we
				currently have.
			*)
			if (err = noErr) and mySFReply.sfGood then begin
				cpb.ioVRefNum := mySFReply.sfFile.vRefNum;
				cpb.ioDirID := mySFReply.sfFile.parID;
				cpb.ioNamePtr := @mySFReply.sfFile.name;
				cpb.ioFDirIndex := -1;
				err := PBGetCatInfoSync(@cpb);
			end; (* if *)
		end; (* if *)
	end; (* StandardGetFolder *)

	function DoStandardGetFolder(var chosenFolderSpec : FSSpec) : OSStatus;
		var
			err : OSStatus;
			newReply: StandardFileReply;
	begin
		StandardGetFolder('', newReply);
		if newReply.sfGood then begin
			chosenFolderSpec := newReply.sfFile;
			err := noErr;
		end else begin
			err := userCanceledErr;
		end; (* if *)
		DoStandardGetFolder := err;
	end; (* DoStandardGetFolder *)

	function DoStandardGetFile (numberOfTypes : integer; typeListPtr: ConstSFTypeListPtr; 
									var chosenFileSpec: FSSpec): OSStatus;
		var
			err: OSStatus;
			newReply: StandardFileReply;
	begin
		err := noErr;
		StandardGetFile(nil, numberOfTypes, typeListPtr, newReply);
		if newReply.sfGood then begin
			chosenFileSpec := newReply.sfFile;
		end else begin
			err := userCanceledErr;
		end; (* if *)
		
		DoStandardGetFile := err;
	end; (* DoStandardGetFile *)

	function DoStandardPutFile(prompt : Str255; name: Str255;
								var chosenFileSpec: FSSpec; var cookie : Ptr) : OSStatus;
		(* See comment in interface part. *)
		var
			err: OSStatus;
			newReply: StandardFileReply;
	begin
		cookie := nil;
		(* Call the new standard file routines. *)
		StandardPutFile(prompt, name, newReply);
		if newReply.sfGood then begin
			chosenFileSpec := newReply.sfFile;
			err := noErr;
		end else begin
			err := userCanceledErr;
		end; (* if *)
		DoStandardPutFile := err;
	end; (* DoStandardPutFile *)

	function DoStandardAskSaveChanges(documentName : Str255; quitting : Boolean;
										var saveThem : Boolean) : OSStatus;
		var
			err : OSStatus;
			actionString : Str255;
			hitItem : integer;
	begin
		{$unused quitting}
		err := noErr;
		if quitting then begin
			actionString := GetIndStr(rMiscStrings, strBeforeQuitting);
		end else begin
			actionString := '';
		end; (* if *)
		ParamText(documentName, actionString, '', '');
		InitCursor;
		hitItem := CautionAlert(rSaveChangesAlert, gOKCancelDiscardAlertFilter);
		case hitItem of
			ditOK:
				saveThem := true;
			ditDontSave:
				saveThem := false;
			ditCancel:
				err := userCanceledErr;
			otherwise
				begin
					ICDebugStr('DoStandardAskSaveChanges: Unexpected hitItem');
					err := -1;
				end
		end; (* case *)
		DoStandardAskSaveChanges := err;
	end; (* DoStandardAskSaveChanges *)

{$endc}

	(* ***** StandardGetFolder for Navigation Services ****** *)

	var
		gNavEventUPP : NavEventUPP;
		gNavObjectFilterUPP : NavObjectFilterUPP;

	procedure MyNavEventProc(callBackSelector: NavEventCallbackMessage; callBackParms: NavCBRecPtr; callBackUD: NavCallBackUserData);
		var
			event : EventRecord;
	begin
		{$unused callBackUD}
		case callBackSelector of
			kNavCBEvent:
				begin
					event := callBackParms^.eventData.eventDataParms.event^;
					if (event.what = updateEvt) |
								(event.what = activateEvt) | 
								((event.what = osEvt) & (band(bsr(event.message, 24), $FF) = suspendResumeMessage)) then begin
						HandleOneEvent(event);
					end; (* if *)
				end;
			otherwise
				(* do nothing *) ;
		end; (* case *)
	end; (* MyNavEventProc *)
	
	var
		gNavLoaded : Boolean;
		
	function NavLoadQ : OSStatus;
		var
			err : OSStatus;
	begin
		err := noErr;
		if not gNavLoaded then begin
			err := NavLoad;
			gNavLoaded := (err = noErr);
		end; (* if *)
		NavLoadQ := err;
	end; (* NavLoadQ *)
	
	procedure NavUnloadQ;
		var
			junk : OSStatus;
	begin
		if gNavLoaded then begin
			junk := NavUnload;
			ICAssert(junk = noErr);
		end; (* if *)
	end; (* NavUnloadQ *)
	
	function NavExtractSingleItem(reply : NavReplyRecord; var fss : FSSpec) : OSStatus;
		var
			err : OSStatus;
			{$ifc icDebug}
				itemCount : longint;
			{$endc}
			junkKeyword : AEKeyword;
			junkType : DescType;
			junkSize : Size;
	begin
		ICAssert((AECountItems(reply.selection, itemCount) = noErr) & (itemCount = 1));
		err := AEGetNthPtr(reply.selection, 1, typeFSS, junkKeyword, 
							junkType, @fss, sizeof(fss), junkSize);
		if err = noErr then begin
			ICAssert(junkType = typeFSS);
			ICAssert(junkSize = sizeof(FSSpec));
			ICAssert(length(fss.name) > 0);
		end; (* if *)
		NavExtractSingleItem := err;
	end; (* NavExtractSingleItem *)
	
	function DoNavGetFolder(var chosenFolderSpec : FSSpec) : OSStatus;
		var
			err : OSStatus;
			junk : OSStatus;
			dialogOptions : NavDialogOptions;
			reply : NavReplyRecord;
	begin
		err := NavLoadQ;
		if err = noErr then begin
			err := NavGetDefaultDialogOptions(dialogOptions);
		end; (* if *)
		if err = noErr then begin
			err := NavChooseFolder(nil, reply, dialogOptions, gNavEventUPP, nil, nil);
			if (err = noErr) & reply.validRecord then begin
				err := NavExtractSingleItem(reply, chosenFolderSpec);
				junk := NavDisposeReply(reply);
				ICAssert(junk = noErr);
			end; (* if *)
		end; (* if *)
		DoNavGetFolder := err;
	end; (* DoNavGetFolder *)

	function MyNavObjectFilterProc(var theItem: AEDesc; info: univ Ptr; callBackUD: NavCallBackUserData; filterMode: NavFilterModes) : Boolean;
		var
			result : Boolean;
			requestedType : OSType;
			foundType : OSType;
			theInfo : NavFileOrFolderInfoPtr;
	begin
		{$unused filterMode}
		result := true;
		
		if theItem.descriptorType = typeFSS then begin
			theInfo := NavFileOrFolderInfoPtr(info);
			if not theInfo^.isFolder then begin
				foundType := theInfo^.finderInfo.fdType;
				requestedType := OSType(callBackUD);
				case longint(requestedType) of
					0:
						result := true;
					longint('APPL'):
						result := (foundType = kApplicationAliasType) | IsApplicationType(foundType);
					otherwise
						result := foundType = requestedType;
				end; (* case *)
			end; (* if *)
		end; (* if *)
		MyNavObjectFilterProc := result;
	end; (* MyNavObjectFilterProc *)
	
	function DoNavGetFile(requestedFileType : OSType;
								var chosenFileSpec : FSSpec) : OSStatus;
		var
			err : OSStatus;
			junk : OSStatus;
			dialogOptions : NavDialogOptions;
			reply : NavReplyRecord;
	begin
		err := NavLoadQ;
		if err = noErr then begin
			err := NavGetDefaultDialogOptions(dialogOptions);
		end; (* if *)
		if err = noErr then begin
			dialogOptions.dialogOptionFlags :=
									kNavDefaultNavDlogOptions
									- kNavAllowPreviews
									- kNavAllowMultipleFiles;
									// kNavNoTypePopup +
									// kNavDontAutoTranslate +
									// kNavDontAddTranslateItems
									// kNavAllFilesInPopup
									// kNavAllowStationery +
									// kNavAllowPreviews +
									// kNavAllowMultipleFiles +
									// kNavAllowInvisibleFiles +
									// kNavDontResolveAliases +
									// kNavSelectDefaultLocation +
									// kNavSelectAllReadableItem
			if requestedFileType = ICfiletype then begin
				err := NavGetFile(nil,					// defaultLocation
									reply,				// reply
									dialogOptions,		// dialogOptions
									gNavEventUPP,		// eventProc
									nil,				// previewProc
									gNavObjectFilterUPP,// filterProc
									nil,				// typeList
									NavCallBackUserData(requestedFileType));	// callBackUD
			end else begin
				err := NavChooseFile(nil,				// defaultLocation
									reply,				// reply
									dialogOptions,		// dialogOptions
									gNavEventUPP,		// eventProc
									nil,				// previewProc
									gNavObjectFilterUPP,// filterProc
									nil,				// typeList
									NavCallBackUserData(requestedFileType));	// callBackUD
			end; (* if *)
			if (err = noErr) & reply.validRecord then begin
				err := NavExtractSingleItem(reply, chosenFileSpec);
				junk := NavDisposeReply(reply);
				ICAssert(junk = noErr);
			end; (* if *)
		end; (* if *)
		
		DoNavGetFile := err;
	end; (* DoNavGetFile *)

	function DoNavPutFile(prompt : Str255; name: Str255; fileType, fileCreator : OSType;
								var chosenFileSpec: FSSpec; var cookie : Ptr) : OSStatus;
		var
			err : OSStatus;
			junk : OSStatus;
			needsDispose : Boolean;
			replyPtr : NavReplyRecordPtr;
			dialogOptions : NavDialogOptions;
	begin
		{$unused prompt}
		
		needsDispose := false;
		replyPtr := NavReplyRecordPtr(NewPtr(sizeof(NavReplyRecord)));
		err := CheckMemError(replyPtr);
		if err = noErr then begin
			err := NavLoadQ;
		end; (* if *)
		if err = noErr then begin
			err := NavGetDefaultDialogOptions(dialogOptions);
		end; (* if *)
		if err = noErr then begin
			dialogOptions.dialogOptionFlags := kNavDefaultNavDlogOptions
								+ kNavNoTypePopup
								- kNavAllowStationery
								- kNavAllowPreviews
								- kNavAllowMultipleFiles;
			dialogOptions.savedFileName := name;
			// As a matter of policy, I don't add the promt to the dialog.
			// This is because the old standard file prompt "Save Extension
			// as:" doesn't work very well in the Nav dialog.  One day
			// I'll figure out how to conditionalise this, but for the moment
			// we'll leave it promptless.
			// -- Quinn, IC 2.1d2, 25 Jul 1998
			//
			// dialogOptions.message := prompt;
			err := NavPutFile(nil,
							replyPtr^,
							dialogOptions,
							gNavEventUPP, 
							fileType,
							fileCreator,
							nil);
			if (err = noErr) & replyPtr^.validRecord then begin
				needsDispose := true;
				err := NavExtractSingleItem(replyPtr^, chosenFileSpec);
			end; (* if *)
		end; (* if *)
		
		// Clean up.
		
		if err <> noErr then begin
			if replyPtr <> nil then begin
				if needsDispose then begin
					junk := NavDisposeReply(replyPtr^);
					ICAssert(junk = noErr);
				end; (* if *)
				DisposePtr(Ptr(replyPtr));
				ICAssert(MemError = noErr);
				replyPtr := nil;
			end; (* if *)
		end; (* if *)
		
		cookie := Ptr(replyPtr);
				
		DoNavPutFile := err;
	end; (* DoNavPutFile *)

	function DoNavStandardCompleteSave(cookie : Ptr) : OSStatus;
		var
			err : OSStatus;
			replyPtr : NavReplyRecordPtr;
	begin
		replyPtr := NavReplyRecordPtr(cookie);
		err := NavCompleteSave(replyPtr^, kNavTranslateInPlace);
		DisposePtr(Ptr(replyPtr));
		ICAssert(MemError = noErr);
		DoNavStandardCompleteSave := err;
	end; (* DoNavStandardCompleteSave *)

	function DoNavAskSaveChanges(documentName : Str255; quitting : Boolean;
										var saveThem : Boolean) : OSStatus;
		var
			err : OSStatus;
			dialogOptions : NavDialogOptions;
			action : NavAskSaveChangesAction;
			result : NavAskSaveChangesResult;
	begin
		err := NavLoadQ;
		if err = noErr then begin
			err := NavGetDefaultDialogOptions(dialogOptions);
		end; (* if *)
		if err = noErr then begin
			dialogOptions.clientName := GetIndStr(rMiscStrings, strInternetConfigApplicationName);
			dialogOptions.savedFileName := documentName;
			if quitting then begin
				action := kNavSaveChangesQuittingApplication
			end else begin
				action := kNavSaveChangesOther;
			end; (* if *)
			err := NavAskSaveChanges(dialogOptions, action, result, gNavEventUPP, nil);
			if err = noErr then begin
				case result of
					kNavAskSaveChangesSave:
						saveThem := true;
					kNavAskSaveChangesCancel:
						err := userCanceledErr;
					kNavAskSaveChangesDontSave:
						saveThem := false;
					otherwise
						begin
							ICDebugStr('DoNavAskSaveChanges: Bogus result code');
							err := -1;
						end;
				end; (* case *)
			end; (* if *)
		end; (* if *)
		DoNavAskSaveChanges := err;
	end; (* DoNavAskSaveChanges *)
	
	(* ***** API Routines ***** *)
	
	function UseNav : Boolean;
	begin
		{$ifc TARGET_API_MAC_CARBON}
			UseNav := true;
		{$elsec}
			UseNav := gHasNavServices & not IsKeyDown(kOptionKeyCode);
		{$endc}
	end; (* UseNav *)

	function ICStandardGetFile (requestedFileType: OSType; 
									var chosenFileSpec: FSSpec;
									var infoForChosenFile: FInfo): OSStatus;
		(* See comment in interface part. *)
		var
			err: OSStatus;
			typeList: SFTypeList;
			numberOfTypes: integer;
	begin
		(* Calculate typeList and numberOfTypes based on input parameters. *)
		typeList[0] := requestedFileType;
		if requestedFileType = OSType(0) then begin
			numberOfTypes := -1;
		end else if requestedFileType = 'APPL' then begin
			typeList[1] := kApplicationAliasType;
			typeList[2] := 'APPC';
			typeList[3] := 'appe';
			
			(*	If you add any more, you should change the declaration of
				typeList, because it's defined to only contain 4 entries.
				You should also make a similar change in IsApplicationType
				in "ICCommonSubs.p".
			*)
			
			numberOfTypes := 4;
		end else begin
			numberOfTypes := 1;
		end; (* if *)
		
		if UseNav then begin
			err := DoNavGetFile(requestedFileType, chosenFileSpec);
		end else begin
			{$ifc TARGET_API_MAC_CARBON}
				ICDebugStr('ICStandardGetFile: UseNav returned false');
				err := -1;
			{$elsec}
				err := DoStandardGetFile(numberOfTypes, @typeList, chosenFileSpec);
			{$endc}
		end; (* if *)

		(* Now call new standard file routine. *)
		if err = noErr then begin
			err := FSpGetFInfo(chosenFileSpec, infoForChosenFile);
		end; (* if *)
		
		ICStandardGetFile := err;
	end; (* ICStandardGetFile *)

	function ICStandardPutFile(prompt : Str255; name: Str255; fileType, fileCreator : OSType;
								var chosenFileSpec: FSSpec; var cookie : Ptr): OSStatus;
		(* See comment in interface part. *)
		var
			err : OSStatus;
	begin
		if UseNav then begin
			err := DoNavPutFile(prompt, name, fileType, fileCreator, chosenFileSpec, cookie);
		end else begin
			{$ifc TARGET_API_MAC_CARBON}
				ICDebugStr('ICStandardPutFile: UseNav returned false');
				err := -1;
			{$elsec}
				err := DoStandardPutFile(prompt, name, chosenFileSpec, cookie);
			{$endc}
		end; (* if *)
		ICStandardPutFile := err;
	end; (* ICStandardPutFile *)
	
	function ICStandardCompleteSave(cookie : Ptr) : OSStatus;
		var
			err : OSStatus;
	begin
		err := noErr;
		if cookie <> nil then begin
			err := DoNavStandardCompleteSave(cookie);
		end; (* if *)
		ICStandardCompleteSave := err;
	end; (* ICStandardCompleteSave *)
	
	function ICStandardGetFolder (var chosenFolderSpec: FSSpec;
									var dirIDOfChosenFolder: longint): OSStatus;
		(* See comment in interface part. *)
		var
			err: OSStatus;
			cpb: CInfoPBRec;
	begin
		(* Under System 7, we have a huge routine that implements this, so we
			just call it.
		*)
		if UseNav then begin
			err := DoNavGetFolder(chosenFolderSpec);
		end else begin
			{$ifc TARGET_API_MAC_CARBON}
				ICDebugStr('ICStandardGetFolder: UseNav returned false');
				err := -1;
			{$elsec}
				err := DoStandardGetFolder(chosenFolderSpec);
			{$endc}
		end; (* if *)
		
		(* Now work out dirIDOfChosenFolder. *)
		if err = noErr then begin
			err := FSpGetCatInfo(chosenFolderSpec, -1, cpb);
			if err = noErr then begin
				chosenFolderSpec.parID := cpb.ioDrParID;
				dirIDOfChosenFolder := cpb.ioDirID;
			end; (* if *)
		end; (* if *)
		
		ICStandardGetFolder := err;
	end; (* ICStandardGetFolder *)

	function ICStandardAskSaveChanges(documentName : Str255; quitting : Boolean;
										var saveThem : Boolean) : OSStatus;
		(* See comment in interface part. *)
		var
			err : OSStatus;
	begin
		if UseNav then begin
			err := DoNavAskSaveChanges(documentName, quitting, saveThem);
		end else begin
			{$ifc TARGET_API_MAC_CARBON}
				ICDebugStr('ICStandardAskSaveChanges: UseNav returned false');
				err := -1;
			{$elsec}
				err := DoStandardAskSaveChanges(documentName, quitting, saveThem);
			{$endc}
		end; (* if *)
		ICStandardAskSaveChanges := err;
	end; (* ICStandardAskSaveChanges *)

	procedure InitICStandardFile;
	begin
		{$ifc not TARGET_API_MAC_CARBON}
			gStandardGetFolderFileFilterUPP := NewFileFilterYDUPP(@StandardGetFolderFileFilter);
			ICAssert(gStandardGetFolderFileFilterUPP <> nil);
			gStandardGetFolderDialogUPP := NewDlgHookYDUPP(@StandardGetFolderDialogHook);
			ICAssert(gStandardGetFolderDialogUPP <> nil);
		{$endc}
		gNavEventUPP := NewNavEventUPP(@MyNavEventProc);
		ICAssert(gNavEventUPP <> nil);
		gNavObjectFilterUPP := NewNavObjectFilterUPP(@MyNavObjectFilterProc);
		ICAssert(gNavObjectFilterUPP <> nil);
	end; (* InitICStandardFile *)
	
	procedure TermICStandardFile;
	begin
		NavUnloadQ;
	end; (* TermICStandardFile *)

end. (* ICStandardFile *)
