/*	This file is part of the Internet Configuration system and is placed in the public domain for the benefit of all.

	This module holds the C implementation of a simple resource fork sanity checker.
	The code is taken from the Disinfectant source code and was contributed by John Norstad.

	I added pragmas so that I could build with full warnings and yet ignore some of
	the warnings that the MPW C compiler generates for this code.  I really didn't want
	to mess with the code.  Remember the engineer's maxim -- if it works, don't fuck with it.
	
	Hmmm, except that I found a couple of cases where it didn't work, so I had
	to mess with it a bit.  Specifically, I tidied up the error handling
	(Quinn-style handling was too hard to retrofit, so I used standard Apple style;
	it uses gotos but them's the breaks) and fixed up the semantics a bit.
	And lo, I got rid of all the warnings at the same time.
	-- Quinn, IC 2.1d2
	
	For comments on the semantics, see "ICResourceForkSanity.p".
*/

#include <Types.h>
#include <Files.h>
#include <Errors.h>
#include <Memory.h>

#include "ICDebug.h"

pascal OSErr CheckResourceForkSanity (const FSSpec *fss, Boolean *sane);

pascal OSErr CheckResourceForkSanity (const FSSpec *fss, Boolean *sane)
{
	short				refNum;			/* file refnum */
	long				count;			/* number of bytes to read */
	long				logEOF;			/* logical EOF */
	unsigned char		*map;			/* pointer to resource map */
	unsigned long		dataLWA;		/* offset in file of data end */
	unsigned long		mapLWA;			/* offset in file of map end */
	unsigned short		typeFWA;		/* offset from map begin to type list */
	unsigned short		nameFWA;		/* offset from map begin to name list */
	unsigned char		*pType;			/* pointer into type list */
	unsigned char		*pName;			/* pointer to start of name list */
	unsigned char		*pMapEnd;		/* pointer to end of map */
	short				nType;			/* number of resource types in map */
	unsigned char		*pTypeEnd;		/* pointer to end of type list */
	short				nRes;				/* number of resources of given type */
	unsigned short		refFWA;			/* offset from type list to ref list */
	unsigned char		*pRef;			/* pointer into reference list */
	unsigned char		*pRefEnd;		/* pointer to end of reference list */
	unsigned short		resNameFWA;		/* offset from name list to resource name */
	unsigned char		*pResName;		/* pointer to resource name */
	unsigned long		resDataFWA;		/* offset from data begin to resource data */
	Boolean				mapOK;			/* true if map is sane */
	OSErr				err;			/* error code */
	
	struct {
		unsigned long		dataFWA;	/* offset in file of data */
		unsigned long		mapFWA;		/* offset in file of map */
		unsigned long		dataLen;	/* data area length */
		unsigned long		mapLen;		/* map area length */
	} header;
	
	ICAssert(fss != nil);
	ICAssert(sane != nil);
	
	// Prepare for failure.
	
	map = nil;
	refNum = 0;
	*sane = true;
	
	/* Open the resource file. */
	
	err = FSpOpenRF(fss, fsRdPerm, &refNum);
	if (err != noErr) goto errorExit;
	
	/* Get the logical eof of the file. */
	
	err = GetEOF(refNum, &logEOF);
	if (err != noErr) goto errorExit;
	if (logEOF == 0) goto errorExit;
	
	/* Read and validate the resource header. */
	
	count = 16;
	err = FSRead(refNum, &count, (Ptr)&header);
	if (err != noErr) goto errorExit;

	dataLWA = header.dataFWA + header.dataLen;
	mapLWA = header.mapFWA + header.mapLen;
	mapOK = count == 16 && header.mapLen > 28 &&
		header.dataFWA < 0x01000000 && header.mapFWA < 0x01000000 &&
		dataLWA <= logEOF && mapLWA <= logEOF &&
		(dataLWA <= header.mapFWA || mapLWA <= header.dataFWA);
	if ( !mapOK ) goto insanityExit;
	
	/* Read the resource map. */
	
	map = (unsigned char *) NewPtr(header.mapLen);
	err = MemError();
	if (err != noErr) goto errorExit;

	err = SetFPos(refNum, fsFromStart, header.mapFWA);
	if (err != noErr) goto errorExit;
	
	count = header.mapLen;
	err = FSRead(refNum, &count, map);
	if (err != noErr) goto errorExit;
	
	/* Verify the type list and name list offsets. */
	
	typeFWA = *(unsigned short*)(map+24);
	nameFWA = *(unsigned short*)(map+26);
	mapOK = typeFWA == 28 && nameFWA >= typeFWA && nameFWA <= header.mapLen &&
		!(typeFWA & 1) && !(nameFWA & 1);
	if ( !mapOK ) goto insanityExit;
	
	/* Verify the type list, reference lists, and name list. */
	
	pType = map + typeFWA;
	pName = map + nameFWA;
	pMapEnd = map + header.mapLen;
	nType = *(short*)pType + 1;
	pType += 2;
	pTypeEnd = pType + (nType<<3);
	mapOK = pTypeEnd <= pMapEnd;
	if ( !mapOK ) goto insanityExit;

	while (pType < pTypeEnd) {
		nRes = *(short*)(pType+4) + 1;
		refFWA = *(unsigned short*)(pType+6);
		pRef = map + typeFWA + refFWA;
		pRefEnd = pRef + 12*nRes;
		if (!(pRef >= pTypeEnd && pRef < pName && !(refFWA & 1))) goto insanityExit;
		while (pRef < pRefEnd) {
			resNameFWA = *(unsigned short*)(pRef+2);
			if (resNameFWA != 0xFFFF) {
				pResName = pName + resNameFWA;
				if (!(pResName + *pResName < pMapEnd)) goto insanityExit;
			};
			resDataFWA = *(unsigned long*)(pRef+4) & 0x00FFFFFF;
			if (!(header.dataFWA + resDataFWA < dataLWA)) goto insanityExit;
			pRef += 12;
		};
		pType += 8;
	};
	err = noErr;				// not necessary, but it makes the next line clearer
	goto errorExit;
	
insanityExit:
	*sane = false;
	
errorExit:

	/* Dispose of the resource map, close the file and return. */
	
	if (map != nil) {
		DisposePtr( (Ptr) map);
		ICAssert(MemError() == noErr);
	}
	if (refNum != 0) {
		OSErr junk;
		
		junk = FSClose(refNum);
		ICAssert(junk == noErr);
	}
	return err;
}
