unit ICMiscSubs;

(*	This file is part of the Internet Configuration system and is placed in the public domain for the benefit of all.

	This file holds all those miscellaneous little functions that are basically wrappers
	around existing OS functionality.  The focus here is Toolbox-level stuff.
*)

interface

	uses
		Files,
		MacWindows,
		Lists, 
		AppleEvents, 

		InternetConfig;

	(* ***** QuickDraw Stuff ***** *)

	procedure DrawIcon (resourceID: integer; const iconRect: Rect; drawHighlighted: Boolean);
		(* This routine draws in icon from the resources specified by resourceID.
			If the System 7 icon utilities are available, it uses the icon family
			resources 'icl8', and draws using the icon utilities.  If theyre not available,
			it uses the 'ICN#' resource and draws using PlotIcon.
		*)
	
	procedure MagicMarkerMode;
		(* This routine sets the HiliteMode low memory global such that the
			next invert operation is done using the user specified highlight colour.
			If Colour QuickDraw isnt available, it does nothing.
		*)

	procedure SetWatchCursor;
		(* Set the cursor to the standard "watch" cursor. *)
		
	function GetScreenBitsBounds : Rect;
	
	procedure DrawGreyBox(const itemRect : Rect);

	(* ***** Event Manager Stuff ***** *)

	function DirtyKey (typedChar: char): Boolean;
		(* This function returns true if the given character will cause a Text
			Edit field to become dirty, ie its a character that will go into
			the field rather than move the insertion point.
		*)
		
	function IsKeyDown (keyCode: integer): Boolean;
		(* Returns true if the given virtual key is down. *)

	(* ***** Window Manager Stuff ***** *)

	(* EnterWindow, ExitWindow and the SavedWindowState type are used to implement
		a standard mechanism for saving and restoring window information.  You call
		EnterWindow when you want to work on a window.  This sets up the parameters
		you need and saves the old parameters in the SavedWindowState variable.
		You then call ExitWindow to restore that state.
	*)
	
	type
		SavedWindowInfo =
			record
				oldPort: GrafPtr;
				thisPort: GrafPtr;
				font: integer;
				size: integer;
				face: Style;
			end;

	procedure EnterWindow (window: WindowPtr; font, size: integer; face: Style;
												var saved: SavedWindowInfo);
		(* Set thePort to window and establish the various window state parameters.
			Save the old parameters in saved.
		*)
		
	procedure ExitWindow (const saved: SavedWindowInfo);
		(* Recover the window parameters from saved. *)

	function MyGetWindowContentRegion (theWindow: WindowPtr): RgnHandle;
		(* Returns the windows content region. This is the region currently
			being used, not a copy.  Do not munge it!
		*)
		
	function MyGetWindowStructureRegion (theWindow: WindowPtr): RgnHandle;
		(* Returns the windows structure region. This is the region currently
			being used, not a copy.  Do not munge it!
		*)

	function MyGetWindowVisibleRegion (theWindow: WindowPtr): RgnHandle;

	function TitleBarOnScreen (theWindow: WindowPtr): Boolean;
		(* Returns true if the windows title bar is on the screen.
			Note that this routine only works if the window is visible,
			ie you have called ShowWindow on it.  The standard mechanism
			for using this routine is to ShowWindow the window, then
			call TitleBarOnScreen.  If it returns true, everything is cool.
			Otherwise the window is completely off the screen, so you can
			move it back on without causing visible effects.
		*)
		
	procedure GetWindowRect (theWindow: WindowPtr; var windowRect: Rect);
		(* This routine sets windowRect to the global co-ordinates of 
			the position of the window.  Its typically used for saving window
			state.
		*)
		
	(* ***** Menu Manager Stuff ***** *)

	procedure SetMenuItemEnable (menuH: MenuHandle; item: integer; enable: Boolean);
		(* Enable the item in the MenuHandle if enable is set, disable it otherwise.
			Youve gotta wonder why this isnt in the operating system!
		*)
		
	function FindMenuItem (menuH: MenuHandle; itemTextToSearchFor: Str255; 
												var indexOfItemFound: integer): Boolean;
		(* This routine searches through the Menu Handle looking for
			itemTextToSearchFor.  If it finds it, it returns true and sets
			indexOfItemFound to the position of the matching menu item.
		*)

	procedure SafeAppendMenu (menuH: MenuHandle; itemText: Str255);
		(* The system AppendMenu interprets the new menu item in strange
			ways, attempting to glean command key and other information from 
			the text.  This is obviously bad for things like the items on the 
			Archie menu.  This safe version of the AppendMenu routine
			sets the text window interpreting it.
		*)

	(* ***** List Manager Stuff ***** *)

	(*	The following are replacements for the standard List Manager routines which
		set the background to white before calling through to the List Manager.
		This allows us to have a white list on a window which has a grey background.
	*)
	
	procedure LSetSelectWhite(setIt : Boolean; theCell : Cell; lHandle : ListHandle);
	procedure LAutoScrollWhite(lHandle : ListHandle);
	procedure LScrollWhite(dCols : integer; dRows : integer; lHandle : ListHandle);
	procedure LUpdateWhite(theRgn : RgnHandle; lHandle : ListHandle);
	procedure LDelRowWhite(count : integer; rowNum : integer; lHandle : ListHandle);
	function  LAddRowWhite(count : integer; rowNum : integer; lHandle : ListHandle) : integer;
	function  LClickWhite(pt : Point; modifiers : integer; lHandle : ListHandle) : Boolean;
	procedure LActivateWhite(act : Boolean; lHandle : ListHandle);

	(* All of these List Manager routines are really targetted at one dimensional
		vertical lists.  They dont work well for two dimensional or horizontal
		lists.
	*)
	
	procedure LSetNoSelection (listH: ListHandle);
		(* This routine clears any selection in the list. *)

	procedure LSelectAll(listH: ListHandle);
		(* This routine selects the entire contents of the list. *)
				
	procedure LSetSingleSelection (listH: ListHandle; row: integer);
		(* This routine selects the single cell (0, row) in the list. *)

	(* The LDoKey routine takes a procedural parameter that is uses to fetch
		the text associated with an item in the list so that it can implement
		its "select by typing" function.
	*)
	type
		GetListCellTextProcType = procedure(listH: ListHandle; listCell: Cell; var cellText: Str255);

	procedure LDoKey(listH: ListHandle; var event:EventRecord; getCellText: GetListCellTextProcType);
		(* This routine processes a key event associated with a list, including
			"select by typing".  You can disable this function by passing nil to
			getCellText.
		*)

	function LSelectedLine (listH: ListHandle): integer;
		(* This function returns the vertical position of the first selected
			cell in the list, or -1 if there is no selected cell.
		*)

	function LIsEmpty (listH: ListHandle): Boolean;
		(* This function returns true if the list is empty. *)
		
	(* ***** Truly Misc Stuff ***** *)

	function AEGotRequiredParams (const theAppleEvent: AppleEvent): OSStatus;
		(* Returns no error if youve extracted all of the required
			parameters out of the AppleEvent.
		*)

	function AEGetParamHandle(const appleEvent: AppleEvent; theAEKeyword: AEKeyword; typeCode: DescType; data : Handle) : OSStatus;

	function HasAppearance : Boolean;
		(* Returns true if Appearance 1.0 APIs are available. *)

	procedure InitICMiscSubs;
		(* The LDoKey function requires a bunch of global state to implement
			its "select by typing" function. This routine initialises that
			information.
		*)
	
implementation

	uses
		Icons, 
		Errors, 
		Resources, 
		Dialogs, 
		ToolUtils, 
		Traps, 
		LowMem,
		GestaltEqu,
		StringCompare,
		Appearance,
		{$IFC NOT TARGET_API_MAC_CARBON} ICAccessors, {$ENDC}

		InternetConfig,
		
		ICDebug,
		ICCommonSubs,
		ICThemeState;

	(* ***** QuickDraw Stuff ***** *)

	procedure DrawIcon (resourceID: integer; const iconRect: Rect; drawHighlighted: Boolean);
		(* See comment in interface part. *)
		var
			junk: OSStatus;
			iconSuite: Handle;
			transform: integer;
			iconH: Handle;
			tmpIconRect : Rect;
	begin
		tmpIconRect := iconRect;
		if GetIconSuite(iconSuite, resourceID, svAllLargeData) = noErr then begin
			if drawHighlighted then begin
				transform := ttSelected;
			end else begin
				transform := ttNone;
			end; (* if *)
			junk := PlotIconSuite(tmpIconRect, 0, transform, iconSuite);
			junk := DisposeIconSuite(iconSuite, false);
		end else begin
			iconH := Get1Resource('ICN#', resourceID);
			if iconH <> nil then begin
				PlotIcon(tmpIconRect, iconH);
			end; (* if *)
		end; (* if *)
	end; (* DrawIcon *)

	procedure MagicMarkerMode;
		(* See comment in interface part. *)
		var
			hasColourQD : Boolean;
			response : longint;
	begin
		hasColourQD := (Gestalt(gestaltQuickdrawVersion, response) = noErr) &
						(response >= gestalt8BitQD);
		if hasColourQD then begin
			LMSetHiliteMode(band(LMGetHiliteMode, $7F));
		end; (* if *)
	end; (* MagicMarkerMode *)

	procedure SetWatchCursor;
		(* See comment in interface part. *)
		var
			cursorH : CursHandle;
	begin
		cursorH := GetCursor(watchCursor);
		if (cursorH <> nil) & (cursorH^ <> nil) then begin
			SetCursor(cursorH^^);
		end; (* if *)
	end; (* SetWatchCursor *)

	function GetScreenBitsBounds : Rect;
		var
			junk : BitMapPtr;
			screenBits: BitMap;
	begin
		junk := GetQDGlobalsScreenBits(screenBits);
		GetScreenBitsBounds := screenBits.bounds;
	end; (* GetScreenBitsBounds *)

	{$ifc TARGET_API_MAC_CARBON}
	
		procedure DrawGreyBox(const itemRect : Rect);
		begin
			{$unused itemRect}
			ICDebugStr('DrawGreyBox: If you have Carbon (and hence Appearance) why are you calling this?');
		end; (* DrawGreyBox *)
	
	{$elsec}
	
		procedure DrawGreyBox(const itemRect : Rect);
			var
				oldPenState : PenState;
				tmpRect : Rect;
		begin
			tmpRect := itemRect;
			GetPenState(oldPenState);
			PenNormal;
			PenPat(qd.gray);
			FrameRect(tmpRect);
			SetPenState(oldPenState);
		end; (* DrawGreyBox *)

	{$endc}

	(* ***** Event Manager Stuff ***** *)

	function DirtyKey (typedChar: char): Boolean;
		(* See comment in interface part. *)
	begin
		DirtyKey := not(typedChar in [kHomeChar, kEndChar, kHelpChar, kPageUpChar, kPageDownChar,
								kLeftArrowChar, kRightArrowChar, kUpArrowChar, kDownArrowChar]);
	end; (* DirtyKey *)

	function IsKeyDown (keyCode: integer): Boolean;
		(* See comment in interface part. *)
		var
			currentKeys: KeyMap;
	begin
		GetKeys(currentKeys);
		IsKeyDown := currentKeys[keyCode];
	end; (* IsKeyDown *)

	(* ***** Window Manager Stuff ***** *)

	procedure EnterWindow (window: WindowPtr; font, size: integer; face: Style;
												var saved: SavedWindowInfo);
		(* See comment in interface part. *)
		var
			windowPort : CGrafPtr;
	begin
		GetPort(saved.oldPort);
		windowPort := GetWindowPort(window);
		SetPort(GrafPtr(windowPort));
		saved.thisPort := GrafPtr(windowPort);
		saved.font := GetPortTextFont(windowPort);
		saved.size := GetPortTextSize(windowPort);
		saved.face := Style(GetPortTextFace(windowPort));
		TextFont(font);
		TextSize(size);
		TextFace(face);
	end; (* EnterWindow *)

	procedure ExitWindow (const saved: SavedWindowInfo);
		(* See comment in interface part. *)
	begin
		SetPort(saved.thisPort);
		TextFont(saved.font);
		TextSize(saved.size);
		TextFace(saved.face);
		SetPort(saved.oldPort);
	end; (* ExitWindow *)
	
	var
		gTmpContentRgn : RgnHandle;
		gTmpStructureRgn : RgnHandle;
		gTmpVisibleRgn : RgnHandle;
	
	function MyGetWindowRegion(window : WindowRef; inRegionCode : WindowRegionCode; ioWinRgn : RgnHandle): OSStatus;
		(*	OK, this is a mess.  We have to handle the following cases:
			o Carbon -- Just call GetWindowRegion
			o PowerPC -- Test for availability of GetWindowRegion, call it if it's available, otherwise use compatibility code.
			o 68K -- Always use compatibility code.
		*)
		var
			err : OSStatus;
			{$ifc not TARGET_API_MAC_CARBON}
				tmpRgn : RgnHandle;
			{$endc}
	begin
		{$ifc TARGET_API_MAC_CARBON}
			err := GetWindowRegion(window, inRegionCode, ioWinRgn);
		{$elsec}
			{$ifc TARGET_CPU_PPC}
				if @GetWindowRegion <> Ptr(kUnresolvedCFragSymbolAddress) then begin
					err := GetWindowRegion(window, inRegionCode, ioWinRgn);
				end else
			{$endc}
				begin
					tmpRgn := nil;
					case inRegionCode of
						kWindowStructureRgn:
							tmpRgn := WindowPeek(window)^.strucRgn;
						kWindowContentRgn:
							tmpRgn := WindowPeek(window)^.contRgn;
						
						otherwise
							; // do nothing
					end; (* case *)
					if tmpRgn = nil then begin
						ICDebugStr('GetWindowRegion: Bogus inRegionCode');
						err := -1;
					end else begin
						CopyRgn(tmpRgn, ioWinRgn);
						err := noErr;
					end; (* if *)
				end; (* if *)
		{$endc}
		MyGetWindowRegion := err;
	end; (* MyGetWindowRegion *)

	function MyGetWindowContentRegion (theWindow: WindowPtr): RgnHandle;
		(* See comment in interface part. *)
		var
			err : OSStatus;
	begin
		ICAssert(gTmpContentRgn <> nil);
		err := MyGetWindowRegion(theWindow, kWindowContentRgn, gTmpContentRgn);
		ICAssert(err = noErr);
		MyGetWindowContentRegion := gTmpContentRgn;
	end; (* MyGetWindowContentRegion *)

	function MyGetWindowStructureRegion (theWindow: WindowPtr): RgnHandle;
		(* See comment in interface part. *)
		var
			err : OSStatus;
	begin
		ICAssert(gTmpStructureRgn <> nil);
		err := MyGetWindowRegion(theWindow, kWindowStructureRgn, gTmpStructureRgn);
		ICAssert(err = noErr);
		MyGetWindowStructureRegion := gTmpStructureRgn;
	end; (* MyGetWindowStructureRegion *)

	function MyGetWindowVisibleRegion (theWindow: WindowPtr): RgnHandle;
		(* See comment in interface part. *)
	begin
		ICAssert(gTmpVisibleRgn <> nil);
		MyGetWindowVisibleRegion := GetPortVisibleRegion(GetWindowPort(theWindow), gTmpVisibleRgn);
	end; (* MyGetWindowVisibleRegion *)

	function TitleBarOnScreen (theWindow: WindowPtr): Boolean;
		(* See comment in interface part. *)
		var
			result : Boolean;
			titleBarRegion: RgnHandle;
	begin
		result := true;
		titleBarRegion := NewRgn;
		if titleBarRegion <> nil then begin
			(* First calculate the title bar region by subtracting the content
				region away from the structure region.
			*)
			CopyRgn(MyGetWindowStructureRegion(theWindow), titleBarRegion);
			DiffRgn(titleBarRegion, MyGetWindowContentRegion(theWindow), titleBarRegion);
			
			(* Now intersect the title bar region with the grey region, ie the region
				describing the extent of the desktop and return true if the intersection
				is not empty.
			*)
			SectRgn(titleBarRegion, GetGrayRgn, titleBarRegion);
			result := not EmptyRgn(titleBarRegion);
			DisposeRgn(titleBarRegion);
		end; (* if *)
		TitleBarOnScreen := result;
	end; (* TitleBarOnScreen *)

	procedure GetWindowRect (theWindow: WindowPtr; var windowRect: Rect);
		(* See comment in interface part. *)
		var
			oldPort : GrafPtr;
			junkRectPtr : RectPtr;
	begin
		GetPort(oldPort);
		SetPort(GrafPtr(GetWindowPort(theWindow)));
		junkRectPtr := GetWindowPortBounds(theWindow, windowRect);
		LocalToGlobal(windowRect.topLeft);
		LocalToGlobal(windowRect.botRight);
		SetPort(oldPort);
	end; (* GetWindowRect *)

	(* ***** Menu Manager Stuff ***** *)

	procedure SetMenuItemEnable (menuH: MenuHandle; item: integer; enable: Boolean);
		(* See comment in interface part. *)
	begin
		if enable then begin
			EnableMenuItem(menuH, item);
		end else begin
			DisableMenuItem(menuH, item);
		end; (* if *)
	end; (* SetMenuItemEnable *)

	function FindMenuItem (menuH: MenuHandle; itemTextToSearchFor: Str255; 
												var indexOfItemFound: integer): Boolean;
		(* See comment in interface part. *)
		var
			itemIndex: integer;
			itemText: Str255;
	begin
		FindMenuItem := false;
		for itemIndex := 1 to CountMenuItems(menuH) do begin
			GetMenuItemText(menuH, itemIndex, itemText);
			if IdenticalString(itemText, itemTextToSearchFor, nil) = 0 then begin
				indexOfItemFound := itemIndex;
				FindMenuItem := true;
			end; (* if *)
		end; (* for *)
	end; (* FindMenuItem *)

	procedure SafeAppendMenu (menuH: MenuHandle; itemText: Str255);
		(* See comment in interface part. *)
	begin
		AppendMenu(menuH, 'fred');

		(*
			If the string begins with a -, we must change it
			before calling SetMenuItemText because the system inteprets
			a leading - as a disabled item, even in SetMenuItemText.
		
			This code is WorldScript safe because the first byte of
			the string is either a) the first byte of a two byte 
			character, in which case it must be high bit set character,
			and - isnt, b) a single byte character, in which case
			the comparison makes sense because all script systems
			contain Roman as the first 128 values of the one byte
			characters.
		*)

		if (length(itemText) > 0) & (itemText[1] = '-') then begin
			itemText := chr(0) + itemText;
		end; (* if *)

		SetMenuItemText(menuH, CountMenuItems(menuH), itemText);
	end; (* SafeAppendMenu *)

	(* ***** List Manager Stuff ***** *)

	procedure LSetSelectWhite(setIt : Boolean; theCell : Cell; lHandle : ListHandle);
		var
			err : OSStatus;
			state : MoreThemeDrawingState;
	begin
		err := MoreGetThemeDrawingState(state);
		if err = noErr then begin
			err := MoreNormalizeThemeDrawingState;
			ICAssert(err = noErr);
			LSetSelect(setIt, theCell, lHandle);
			err := MoreSetThemeDrawingState(state, true);
			ICAssert(err = noErr);
		end; (* if *)
		ICAssert(err = noErr);
	end; (* LSetSelectWhite *)
	
	procedure LAutoScrollWhite(lHandle : ListHandle);
		var
			err : OSStatus;
			state : MoreThemeDrawingState;
	begin
		err := MoreGetThemeDrawingState(state);
		if err = noErr then begin
			err := MoreNormalizeThemeDrawingState;
			ICAssert(err = noErr);
			LAutoScroll(lHandle);
			err := MoreSetThemeDrawingState(state, true);
		end; (* if *)
		ICAssert(err = noErr);
	end; (* LAutoScrollWhite *)
	
	procedure LScrollWhite(dCols : integer; dRows : integer; lHandle : ListHandle);
		var
			err : OSStatus;
			state : MoreThemeDrawingState;
	begin
		err := MoreGetThemeDrawingState(state);
		if err = noErr then begin
			err := MoreNormalizeThemeDrawingState;
			ICAssert(err = noErr);
			LScroll(dCols, dRows, lHandle);
			err := MoreSetThemeDrawingState(state, true);
		end; (* if *)
		ICAssert(err = noErr);
	end; (* LScrollWhite *)
	
	procedure LUpdateWhite(theRgn : RgnHandle; lHandle : ListHandle);
		var
			err : OSStatus;
			state : MoreThemeDrawingState;
	begin
		err := MoreGetThemeDrawingState(state);
		if err = noErr then begin
			err := MoreNormalizeThemeDrawingState;
			ICAssert(err = noErr);
			LUpdate(theRgn, lHandle);
			err := MoreSetThemeDrawingState(state, true);
		end; (* if *)
		ICAssert(err = noErr);
	end; (* LUpdateWhite *)
	
	procedure LDelRowWhite(count : integer; rowNum : integer; lHandle : ListHandle);
		var
			err : OSStatus;
			state : MoreThemeDrawingState;
	begin
		err := MoreGetThemeDrawingState(state);
		if err = noErr then begin
			err := MoreNormalizeThemeDrawingState;
			ICAssert(err = noErr);
			LDelRow(count, rowNum, lHandle);
			err := MoreSetThemeDrawingState(state, true);
		end; (* if *)
		ICAssert(err = noErr);
	end; (* LDelRowWhite *)
	
	function  LAddRowWhite(count : integer; rowNum : integer; lHandle : ListHandle) : integer;
		var
			err : OSStatus;
			state : MoreThemeDrawingState;
			result : integer;
	begin
		err := MoreGetThemeDrawingState(state);
		if err = noErr then begin
			err := MoreNormalizeThemeDrawingState;
			ICAssert(err = noErr);
			result := LAddRow(count, rowNum, lHandle);
			err := MoreSetThemeDrawingState(state, true);
		end; (* if *)
		ICAssert(err = noErr);
		LAddRowWhite := result;
	end; (* LAddRowWhite *)
	
	function  LClickWhite(pt : Point; modifiers : integer; lHandle : ListHandle) : Boolean;
		var
			err : OSStatus;
			state : MoreThemeDrawingState;
			result : Boolean;
	begin
		err := MoreGetThemeDrawingState(state);
		if err = noErr then begin
			err := MoreNormalizeThemeDrawingState;
			ICAssert(err = noErr);
			result := LClick(pt, modifiers, lHandle);
			err := MoreSetThemeDrawingState(state, true);
		end; (* if *)
		ICAssert(err = noErr);
		LClickWhite := result;
	end; (* LClickWhite *)
	
	procedure LActivateWhite(act : Boolean; lHandle : ListHandle);
		var
			err : OSStatus;
			state : MoreThemeDrawingState;
	begin
		err := MoreGetThemeDrawingState(state);
		if err = noErr then begin
			err := MoreNormalizeThemeDrawingState;
			ICAssert(err = noErr);
			LActivate(act, lHandle);
			err := MoreSetThemeDrawingState(state, true);
		end; (* if *)
		ICAssert(err = noErr);
	end; (* LActivateWhite *)
	
	var
		gCharsTypedSoFar: Str255;
		gTimeOfLastCharTyped: longint;
		gListHandleOfLastCharTyped: ListHandle;

	procedure LSetNoSelection (listH: ListHandle);
		(* See comment in interface part. *)
		var
			listCell: Cell;
	begin
		listCell.v := 0;
		listCell.h := 0;
		while LGetSelect(true, listCell, listH) do begin
			LSetSelectWhite(false, listCell, listH);
			listCell.v := listCell.v + 1;
			listCell.h := 0;
		end; (* if *)
	end; (* LSetNoSelection *)
	
	procedure LSelectAll(listH: ListHandle);
		var
			listCell: Cell;
			row: integer;
			bounds : ListBounds;
			junk : ListBoundsPtr;
	begin
		junk := GetListDataBounds(listH, bounds);
		for row := 0 to bounds.bottom - 1 do begin
			listCell.v := row;
			listCell.h := 0;
			LSetSelectWhite(true, listCell, listH);
		end; (* for *)
	end; (* LSelectAll *)

	procedure LSetSingleSelection (listH: ListHandle; row: integer);
		(* See comment in interface part. *)
		var
			listCell: Cell;
	begin
		listCell.h := 0;
		listCell.v := row;
		LSetSelectWhite(true, listCell, listH);
		listCell.v := 0;
		listCell.h := 0;
		while LGetSelect(true, listCell, listH) do begin
			if listCell.v <> row then begin
				LSetSelectWhite(false, listCell, listH);
			end; (* if *)
			listCell.v := listCell.v + 1;
			listCell.h := 0;
		end; (* while *)
		LAutoScrollWhite(listH);
	end; (* LSetSingleSelection *)

	function LGetUniqueEntryName (listH: ListHandle; listCell: Cell; getCellText: GetListCellTextProcType): Str255;
		(* This function calls getCellText and then returns a uniquified version of the
			cell text.  What that means is that it returns the cell text followed by
			a chr(0) followed by the the vertical co-ordinate of the cell encoded
			as two characters.  This is useful because it allows functions that
			need to distinguish between two cells even if they have the same
			name to function, eg tabbing.
		*)
		var
			result: Str255;
	begin
		result := '';
		getCellText(listH, listCell, result);
		LGetUniqueEntryName := concat(result, chr(0), 
												chr(listCell.v div 256), 
												chr(listCell.v mod 256));
	end; (* LGetUniqueEntryName *)

	function LGetSelectedCellCommon (listH: ListHandle; var listCell: Cell; getCellText: GetListCellTextProcType;
															first : Boolean): Boolean;
		(* This function finds the alphabetically first or last cell (depending
			on the value of first) in the currently selected cells of the list.  It
			returns false if there are no selected cells.
		*)
		var
			result : Boolean;
			cellText: Str255;
			bestText : Str255;
			indexOfBestText: integer;
	begin
		(* Establish some pre-conditions. *)
		result := false;
		listCell.h := 0;
		listCell.v := 0;
		indexOfBestText := 0;
		if first then begin
			bestText := concat(chr(255), chr(255));
		end else begin
			bestText := '';
		end; (* if *)
		
		(* Loop through the selected cells, looking for the best text (ie the
			alphabetically first or last).
		*)
		while LGetSelect(true, listCell, listH) do begin
			result := true;
			getCellText(listH, listCell, cellText);
			if (first & (CompareString(cellText, bestText, nil) < 0)) |
						(not first & (CompareString(cellText, bestText, nil) > 0)) then begin
				indexOfBestText := listCell.v;
				bestText := cellText;
			end; (* if *)
			listCell.v := listCell.v + 1;
		end; (* while *)
		
		(* Finish up. *)
		listCell.h := 0;
		listCell.v := indexOfBestText;
		LGetSelectedCellCommon := result;
	end; (* LGetSelectedCellCommon *)

	function LGetFirstSelectedCell (listH: ListHandle; var listCell: Cell; getCellText: GetListCellTextProcType): Boolean;
		(* This function finds the alphabetically first cell in the currently
			selected cells of the list.  It returns false if there are no selected cells.
		*)
	begin
		LGetFirstSelectedCell := LGetSelectedCellCommon(listH, listCell, getCellText, true);
	end; (* LGetFirstSelectedCell *)

	function LGetLastSelectedCell (listH: ListHandle; var listCell: Cell; getCellText: GetListCellTextProcType): Boolean;
		(* This function finds the alphabetically last cell in the currently
			selected cells of the list.  It returns false if there are no selected cells.
		*)
	begin
		LGetLastSelectedCell := LGetSelectedCellCommon(listH, listCell, getCellText, false);
	end; (* LGetLastSelectedCell *)

	function LSelectFirstCommon (listH: ListHandle; markerText: Str255; getCellText: GetListCellTextProcType;
													before : Boolean; orEqual : Boolean): Boolean;
		(* This function selects the first cell alphabetically before (or after, depending
			on the value of "before") the markerText.  If returns true if it managed to do this,
			false otherwise.  The orEqual value determines whether an
			equal value is considered to be before the otherwise best value.
		*)
		var
			result: Boolean;
			row : integer;
			indexOfBestText: integer;
			listCell: Cell;
			bestText : Str255;
			cellText: Str255;
			comp1 : integer;
			comp2 : integer;
			bounds : ListBounds;
			junk : ListBoundsPtr;
	begin
		(* Establish some pre-conditions. *)
		result := false;
		indexOfBestText := 0;
		if before then begin
			bestText := '';
		end else begin
			bestText := concat(chr(255), chr(255));
		end; (* if *)
		
		(* Iterate through all the cells, looking for best text.  Best is defined
			as the phone thats alphabetically before (or after, depending on
			the value of before) the markerText.
		*)
		junk := GetListDataBounds(listH, bounds);
		for row := 0 to bounds.bottom - 1 do begin
			listCell.h := 0;
			listCell.v := row;
			getCellText(listH, listCell, cellText);

			(* OK, so this needs some explaining (-:
				comp1 and comp2 just cache the value of the comparisons between
				markerText, cellText and bestText.
				
				If before is true, were looking for the cell immediately before
				the markerText.  This means that the markerText must be
				greater than (ie "comp1 > 0") or equal to (ie "| (comp1 = 0)")
				the cellText, and the cellText must be greater than (ie "comp2 > 0")
				the bestText weve found so far.
				
				If before is false, were looking for the cell immediately after
				the markerText.  This means that the markerText must be
				less than (ie "comp1 < 0") or equal to (ie "| (comp1 = 0)")
				the cellText, and the cellText must be less than (ie "comp2 < 0")
				the bestText weve found so far.
				
				*phew*
			*)
			comp1 := CompareString(markerText, cellText, nil);
			comp2 := CompareString(cellText, bestText, nil);
			if (		before & (((comp1 > 0) | ((comp1 = 0) & orEqual)) & (comp2 > 0))) |
				(not before & (((comp1 < 0) | ((comp1 = 0) & orEqual)) & (comp2 < 0))) then begin
				bestText := cellText;
				indexOfBestText := listCell.v;
				result := true;
			end; (* if *)
		end; (* for *)
		
		(* Now set the selection to the cell we found. *)
		if result then begin
			LSetSingleSelection(listH, indexOfBestText);
		end; (* if *)
		LSelectFirstCommon := result;
	end; (* LSelectFirstCommon *)

	function LSelectFirstBefore (listH: ListHandle; beforeThis: Str255; getCellText: GetListCellTextProcType): Boolean;
		(* This function selects the first cell alphabetically before
			the beforeThis text.  If returns true if it managed to do this,
			false otherwise.  The orEqual value determines whether an
			equal value is considered to be before the otherwise best value.
		*)
	begin
		LSelectFirstBefore := LSelectFirstCommon(listH, beforeThis, getCellText, true, false);
	end; (* LSelectFirstBefore *)
	
	function LSelectFirstAfter (listH: ListHandle; afterThis: Str255; getCellText: GetListCellTextProcType; orEqual:Boolean): Boolean;
		(* This function selects the first cell alphabetically after
			the afterThis text.  If returns true if it managed to do this,
			false otherwise.  The orEqual value determines whether an
			equal value is considered to be before the otherwise best value.
		*)
	begin
		LSelectFirstAfter := LSelectFirstCommon(listH, afterThis, getCellText, false, orEqual);
	end; (* LSelectFirstAfter *)

	procedure LDownArrow(listH : ListHandle);
		(* Find the last selected cell and select the cell after it. *)
		var
			listCell : Cell;
			indexOfCellToSelect : integer;
			bounds : ListBounds;
			junk : ListBoundsPtr;
	begin
		listCell.h := 0;
		listCell.v := 0;
		indexOfCellToSelect := 0;
		while LGetSelect(true, listCell, listH) do begin
			listCell.v := listCell.v + 1;
			indexOfCellToSelect := listCell.v;
		end; (* if *)
		junk := GetListDataBounds(listH, bounds);
		if indexOfCellToSelect >= bounds.bottom then begin
			indexOfCellToSelect := bounds.bottom - 1;
		end; (* if *)
		LSetSingleSelection(listH, indexOfCellToSelect);
		LAutoScrollWhite(listH);
	end; (* LDownArrow *)

	procedure LUpArrow(listH : ListHandle);
		(* Find the first selected cell and select the cell before it. *)
		var
			listCell : Cell;
			bounds : ListBounds;
			junk : ListBoundsPtr;
	begin
		listCell.h := 0;
		listCell.v := 0;
		junk := GetListDataBounds(listH, bounds);
		if not LGetSelect(true, listCell, listH) then begin
			listCell.v := bounds.bottom;
		end; (* if *)
		if listCell.v > 0 then begin
			listCell.v := listCell.v - 1;
		end; (* if *)
		LSetSingleSelection(listH, listCell.v);
		LAutoScrollWhite(listH);
	end; (* LUpArrow *)
	
	procedure LTab(listH : ListHandle; getCellText: GetListCellTextProcType; shift : Boolean);
		(* Handle Tab and shift-Tab keys in the list. *)
		var
			junkBool : Boolean;
			listCell : Cell;
			done : Boolean;
			selectedCellText : Str255;
	begin
		if getCellText <> nil then begin
			if not shift then begin
				(* Tab -- If there are selected cells then attempt to select the first
					cell after the last selected cell.  If we cant or there were no 
					selected cells, then select the first cell alphabetically.
				*)
				done := false;
				if LGetLastSelectedCell(listH, listCell, getCellText) then begin
					selectedCellText := LGetUniqueEntryName(listH, listCell, getCellText);
					if LSelectFirstAfter(listH, selectedCellText, getCellText, false) then begin
						done := true;
					end; (* if *)
				end; (* if *)
				if not done then begin
					junkBool := LSelectFirstAfter(listH, '', getCellText, false);
				end; (* if *)
			end else begin
				(* shift-Tab -- If there are no selected cells then attempt to select the
					cell before the first selected cell.  If we cant or there were no 
					selected cells, then select the last cell alphabetically.
				*)
				done := false;
				if LGetFirstSelectedCell(listH, listCell, getCellText) then begin
					getCellText(listH, listCell, selectedCellText);
					if LSelectFirstBefore(listH, selectedCellText, getCellText) then begin
						done := true;
					end; (* if *)
				end; (* if *)
				if not done then begin
					junkBool := LSelectFirstBefore(listH, chr(255), getCellText);
				end; (* if *)
			end; (* if *)
		end; (* if *)
	end; (* LTab *)

	procedure LOtherKey(listH : ListHandle; getCellText : GetListCellTextProcType;
											typedChar : char; eventTicks : longint);
		(* This routine handles the pressing of a normaly key in a list
			by selecting the cell best associated with the text typed so far.
		*)
		var
			junkBool : Boolean;
	begin
		if (getCellText <> nil) & (typedChar >= ' ') then begin
			if eventTicks - gTimeOfLastCharTyped > 60 then begin
				gCharsTypedSoFar := '';
			end; (* if *)
			gTimeOfLastCharTyped := eventTicks;
			gCharsTypedSoFar := concat(gCharsTypedSoFar, typedChar);
			if not LSelectFirstAfter(listH, gCharsTypedSoFar, getCellText, true) then begin
				junkBool := LSelectFirstBefore(listH, chr(255), getCellText);
			end; (* if *)
		end; (* if *)
	end; (* LOtherKey *)

	procedure LDoKey(listH: ListHandle; var event:EventRecord; getCellText: GetListCellTextProcType);
		(* See comment in interface part. *)
		var
			eventTicks: longint;
			typedChar:char;
			bounds : ListBounds;
			visible : ListBounds;
			junk : ListBoundsPtr;
	begin
		eventTicks := event.when;
		typedChar := chr(band(event.message, charCodeMask));

		(* First up, if weve changed lists or typed a control character,
			we reset the globals that track the current typing state.
		*)
		if (gListHandleOfLastCharTyped <> listH) or (typedChar < ' ') then begin
			gTimeOfLastCharTyped := 0;
			gListHandleOfLastCharTyped := listH;
		end; (* if *)
		
		(* Now dispatch the various characters type. *)
		junk := GetListDataBounds(listH, bounds);
		junk := GetListVisibleCells(listH, visible);
		case typedChar of
			(* Handle the trivial scrolling around keys. *)
			kHomeChar:
				LScrollWhite(0, -bounds.bottom, listH);
			kEndChar:
				LScrollWhite(0, bounds.bottom, listH);
			kPageUpChar:
				LScrollWhite(0, -(visible.bottom - visible.top - 2), listH);
			kPageDownChar:
				LScrollWhite(0, (visible.bottom - visible.top - 2), listH);
		
			(* Handle up and down arrows. *)
			kDownArrowChar:
				LDownArrow(listH);
			kUpArrowChar:
				LUpArrow(listH);
				
			(* Tab and shift-Tab and other keys are trickier. *)
			kTabChar:
				LTab(listH, getCellText, band(event.modifiers, shiftKey) <> 0);
			otherwise
				LOtherKey(listH, getCellText, typedChar, eventTicks);
		end; (* case *)
	end; (* LDoKey *)

	function LSelectedLine (listH: ListHandle): integer;
		(* See comment in interface part. *)
		var
			listCell: Cell;
	begin
		SetPt(listCell, 0, 0);
		if LGetSelect(true, listCell, listH) then begin
			LSelectedLine := listCell.v;
		end else begin
			LSelectedLine := -1;
		end; (* if *)
	end; (* LSelectedLine *)

	function LIsEmpty (listH: ListHandle): Boolean;
		(* See comment in interface part. *)
		var
			bounds : ListBounds;
			junk : ListBoundsPtr;
	begin
		junk := GetListDataBounds(listH, bounds);
		LIsEmpty := bounds.bottom <= bounds.top;
	end; (* LIsEmpty *)

	(* ***** Truly Misc Stuff ***** *)

	function AEGotRequiredParams (const theAppleEvent: AppleEvent): OSStatus;
		(* See comment in interface part. *)
		var
			typeCode: DescType;
			actualSize: Size;
			err: OSStatus;
			tmpAppleEvent : AppleEvent;
	begin
		tmpAppleEvent := theAppleEvent;
		err := AEGetAttributePtr(tmpAppleEvent, keyMissedKeywordAttr, typeWildCard, typeCode, nil, 0, actualSize);
		if err = errAEDescNotFound then begin
			err := noErr;
		end else if err = noErr then begin
			err := errAEEventNotHandled;
		end; (* if *)
		AEGotRequiredParams := err;
	end; (* AEGotRequiredParams *)

	function AEGetParamHandle(const theAppleEvent: AppleEvent; theAEKeyword: AEKeyword; typeCode: DescType; data : Handle) : OSStatus;
		var
			err : OSStatus;
			tmpAppleEvent : AppleEvent;
			dataSize : Size;
			s : SInt8;
			actualTypeCode : DescType;
			actualSize : Size;
	begin
		tmpAppleEvent := theAppleEvent;
		err := AESizeOfParam(tmpAppleEvent, theAEKeyword, typeCode, dataSize);
		if err = noErr then begin
			SetHandleSize(data, dataSize);
			err := MemError;
		end; (* if *)
		if err = noErr then begin
			s := HGetState(data);
			HLock(data);
			err := AEGetParamPtr(tmpAppleEvent, theAEKeyword, typeCode, actualTypeCode, data^, GetHandleSize(data), actualSize);
			HSetState(data, s);
		end; (* if *)
		if err = noErr then begin
			ICAssert(actualTypeCode = typeCode);
			ICAssert(actualSize = dataSize);
		end; (* if *)

		AEGetParamHandle := err;
	end; (* AEGetParamHandle *)

	var
		gHasAppearance : Boolean;
		
	function HasAppearance : Boolean;
	begin
		HasAppearance := gHasAppearance;
	end; (* HasAppearance *)

	(* ***** Initialisation/Termination ***** *)
	
	procedure InitICMiscSubs;
		(* See comment in interface part. *)
		var
			gestaltResponse : longint;
	begin
		gCharsTypedSoFar := '';
		gTimeOfLastCharTyped := 0;
		gListHandleOfLastCharTyped := nil;
		
		gTmpContentRgn := NewRgn;
		ICAssert(gTmpContentRgn <> nil);
		gTmpStructureRgn := NewRgn;
		ICAssert(gTmpStructureRgn <> nil);
		gTmpVisibleRgn := NewRgn;
		ICAssert(gTmpVisibleRgn <> nil);
		
		gHasAppearance := (Gestalt(gestaltAppearanceAttr, gestaltResponse) = noErr) &
						btst(gestaltResponse, gestaltAppearanceExists);
	end; (* InitICMiscSubs *)

end. (* ICMiscSubs *)
